# -*- coding: utf-8 -*-
from typing import List, Union, Type, Optional

import pandas as pd

from zvtm.contract import IntervalLevel, TradableEntity, AdjustType
from zvtm.contract.api import get_schema_by_name
from zvtm.contract.factor import Accumulator
from zvtm.domain import Stock
from zvtm.factors.algorithm import live_or_dead
from zvtm.factors.technical_factor import TechnicalFactor
from zvtm.utils import pd_is_not_null


def get_ma_stats_factor_schema(entity_type: str,
                               level: Union[IntervalLevel, str] = IntervalLevel.LEVEL_1DAY):
    if type(level) == str:
        level = IntervalLevel(level)

    schema_str = '{}{}MaStatsFactor'.format(entity_type.capitalize(), level.value.capitalize())

    return get_schema_by_name(schema_str)


class MaStatsAccumulator(Accumulator):
    def __init__(self, acc_window: int = 250, windows=None, vol_windows=None) -> None:
        super().__init__(acc_window)
        self.windows = windows
        self.vol_windows = vol_windows

    def acc_one(self, entity_id, df: pd.DataFrame, acc_df: pd.DataFrame, state: dict) -> (pd.DataFrame, dict):
        self.logger.info(f'acc_one:{entity_id}')
        if pd_is_not_null(acc_df):
            df = df[df.index > acc_df.index[-1]]
            if pd_is_not_null(df):
                self.logger.info(f'compute from {df.iloc[0]["timestamp"]}')
                acc_df = pd.concat([acc_df, df])
            else:
                self.logger.info('no need to compute')
                return acc_df, state
        else:
            acc_df = df

        for window in self.windows:
            col = 'ma{}'.format(window)
            self.indicators.append(col)

            ma_df = acc_df['close'].rolling(window=window, min_periods=window).mean()
            acc_df[col] = ma_df

        acc_df['live'] = (acc_df['ma5'] > acc_df['ma10']).apply(lambda x: live_or_dead(x))
        acc_df['distance'] = (acc_df['ma5'] - acc_df['ma10']) / acc_df['close']

        live = acc_df['live']
        acc_df['count'] = live * (live.groupby((live != live.shift()).cumsum()).cumcount() + 1)

        acc_df['bulk'] = (live != live.shift()).cumsum()
        area_df = acc_df[['distance', 'bulk']]
        acc_df['area'] = area_df.groupby('bulk').cumsum()

        for vol_window in self.vol_windows:
            col = 'vol_ma{}'.format(vol_window)
            self.indicators.append(col)

            vol_ma_df = acc_df['turnover'].rolling(window=vol_window, min_periods=vol_window).mean()
            acc_df[col] = vol_ma_df

        acc_df = acc_df.set_index('timestamp', drop=False)
        return acc_df, state


class MaStatsFactor(TechnicalFactor):

    def __init__(self, entity_schema: Type[TradableEntity] = Stock, provider: str = None, entity_provider: str = None,
                 entity_ids: List[str] = None, exchanges: List[str] = None, codes: List[str] = None,
                 start_timestamp: Union[str, pd.Timestamp] = None, end_timestamp: Union[str, pd.Timestamp] = None,
                 filters: List = None, order: object = None, limit: int = None,
                 level: Union[str, IntervalLevel] = IntervalLevel.LEVEL_1DAY, category_field: str = 'entity_id',
                 time_field: str = 'timestamp', computing_window: int = None, keep_all_timestamp: bool = False,
                 fill_method: str = 'ffill', effective_number: int = None, need_persist: bool = True,
                 only_compute_factor: bool = False, factor_name: str = None, clear_state: bool = False, only_load_factor: bool = False,
                 adjust_type: Union[AdjustType, str] = None, windows=None, vol_windows=None) -> None:
        if need_persist:
            self.factor_schema = get_ma_stats_factor_schema(entity_type=entity_schema.__name__, level=level)

        if not windows:
            windows = [5, 10, 34, 55, 89, 144, 120, 250]
        self.windows = windows

        if not vol_windows:
            vol_windows = [30]
        self.vol_windows = vol_windows

        columns: List = ['id', 'entity_id', 'timestamp', 'level', 'open', 'close', 'high', 'low', 'turnover']

        accumulator: Accumulator = MaStatsAccumulator(windows=self.windows, vol_windows=self.vol_windows)

        super().__init__(entity_schema, provider, entity_provider, entity_ids, exchanges, codes, start_timestamp,
                         end_timestamp, columns, filters, order, limit, level, category_field,
                         time_field, computing_window, keep_all_timestamp, fill_method, effective_number, None,
                         accumulator, need_persist, only_compute_factor, factor_name, clear_state, only_load_factor, adjust_type)


class TFactor(MaStatsFactor):

    def __init__(self, entity_schema: Type[TradableEntity] = Stock, provider: str = None, entity_provider: str = None,
                 entity_ids: List[str] = None, exchanges: List[str] = None, codes: List[str] = None,
                 start_timestamp: Union[str, pd.Timestamp] = None, end_timestamp: Union[str, pd.Timestamp] = None,
                 filters: List = None, order: object = None, limit: int = None,
                 level: Union[str, IntervalLevel] = IntervalLevel.LEVEL_1DAY,
                 category_field: str = 'entity_id', time_field: str = 'timestamp', computing_window: int = None,
                 keep_all_timestamp: bool = False, fill_method: str = 'ffill', effective_number: int = None,
                 need_persist: bool = True, only_compute_factor: bool = False, factor_name: str = None, clear_state: bool = False,
                 only_load_factor: bool = True, adjust_type: Union[AdjustType, str] = None, windows=None,
                 vol_windows=None) -> None:
        super().__init__(entity_schema, provider, entity_provider, entity_ids, exchanges, codes, start_timestamp,
                         end_timestamp, filters, order, limit, level, category_field, time_field, computing_window,
                         keep_all_timestamp, fill_method, effective_number, need_persist, only_compute_factor, factor_name,
                         clear_state, only_load_factor, adjust_type, windows, vol_windows)

    def drawer_sub_df_list(self) -> Optional[List[pd.DataFrame]]:
        return [self.factor_df[['area']]]

    def drawer_factor_df_list(self) -> Optional[List[pd.DataFrame]]:
        return [self.factor_df[['ma5', 'ma10']]]


if __name__ == '__main__':
    from zvtm.factors.ma.domain import *
    codes = ['000338']

    f = TFactor(codes=codes, only_load_factor=False)

    # distribute(f.factor_df[['area']],'area')
    f.draw(show=True)

# the __all__ is generated
__all__ = ['get_ma_stats_factor_schema', 'MaStatsAccumulator', 'MaStatsFactor', 'TFactor']


# -*- coding: utf-8 -*-
from typing import List, Union

import pandas as pd

from zvtm.contract import IntervalLevel
from zvtm.contract.api import get_data, get_db_session
from zvtm.contract.normal_data import NormalData
from zvtm.contract.reader import DataReader
from zvtm.domain import AccountStats, Order, trader_info, TraderInfo, Position
from zvtm.contract.drawer import Drawer


def clear_trader(trader_name, session=None):
    if not session:
        session = get_db_session('zvt', data_schema=TraderInfo)
    session.query(TraderInfo).filter(TraderInfo.trader_name == trader_name).delete()
    session.query(AccountStats).filter(AccountStats.trader_name == trader_name).delete()
    session.query(Position).filter(Position.trader_name == trader_name).delete()
    session.query(Order).filter(Order.trader_name == trader_name).delete()
    session.commit()

def get_trader_info(trader_name=None, return_type='df', start_timestamp=None, end_timestamp=None,
                    filters=None, session=None, order=None, limit=None) -> List[trader_info.TraderInfo]:
    if trader_name:
        if filters:
            filters = filters + [trader_info.TraderInfo.trader_name == trader_name]
        else:
            filters = [trader_info.TraderInfo.trader_name == trader_name]

    return get_data(data_schema=trader_info.TraderInfo, entity_id=None, codes=None, level=None, provider='zvt',
                    columns=None, return_type=return_type, start_timestamp=start_timestamp,
                    end_timestamp=end_timestamp, filters=filters, session=session, order=order, limit=limit)


def get_order_securities(trader_name):
    items = get_db_session(provider='zvt', data_schema=Order).query(Order.entity_id).filter(
        Order.trader_name == trader_name).group_by(Order.entity_id).all()

    return [item[0] for item in items]


class AccountStatsReader(DataReader):

    def __init__(self, start_timestamp: Union[str, pd.Timestamp] = None, end_timestamp: Union[str, pd.Timestamp] = None,
                 columns: List = None, filters: List = None, order: object = None,
                 level: IntervalLevel = IntervalLevel.LEVEL_1DAY, trader_names: List[str] = None) -> None:
        self.trader_names = trader_names

        self.filters = filters

        if self.trader_names:
            filter = [AccountStats.trader_name == name for name in self.trader_names]
            if self.filters:
                self.filters += filter
            else:
                self.filters = filter
        super().__init__(AccountStats, None, None, None, None, None, None,
                         start_timestamp, end_timestamp, columns, self.filters, order, None, level,
                         category_field='trader_name', time_field='timestamp', computing_window=None)

    def draw_line(self, show=True):
        drawer = Drawer(main_data=NormalData(self.data_df.copy()[['trader_name', 'timestamp', 'all_value']],
                                             category_field='trader_name'))
        return drawer.draw_line(show=show)


class OrderReader(DataReader):
    def __init__(self, start_timestamp: Union[str, pd.Timestamp] = None, end_timestamp: Union[str, pd.Timestamp] = None,
                 columns: List = None, filters: List = None, order: object = None, level: IntervalLevel = None,
                 trader_names: List[str] = None) -> None:
        self.trader_names = trader_names

        self.filters = filters

        if self.trader_names:
            filter = [Order.trader_name == name for name in self.trader_names]
            if self.filters:
                self.filters += filter
            else:
                self.filters = filter

        super().__init__(Order, None, None, None, None, None, None,
                         start_timestamp, end_timestamp, columns, self.filters, order, None, level,
                         category_field='trader_name', time_field='timestamp', computing_window=None)


if __name__ == '__main__':
    reader = AccountStatsReader(trader_names=['000338_ma_trader'])
    drawer = Drawer(main_data=NormalData(reader.data_df.copy()[['trader_name', 'timestamp', 'all_value']],
                                         category_field='trader_name'))
    drawer.draw_line()
# the __all__ is generated
__all__ = ['clear_trader', 'get_trader_info', 'get_order_securities', 'AccountStatsReader', 'OrderReader']


# -*- coding: utf-8 -*-
from zvtm.consts import SAMPLE_STOCK_CODES
from zvtm.domain import RightsIssueDetail, DividendFinancing
from zvtm.recorders.eastmoney.common import EastmoneyPageabeDataRecorder
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.utils.time_utils import now_pd_timestamp
from zvtm.utils.utils import to_float


class RightsIssueDetailRecorder(EastmoneyPageabeDataRecorder):
    data_schema = RightsIssueDetail

    url = 'https://emh5.eastmoney.com/api/FenHongRongZi/GetPeiGuMingXiList'
    page_url = url
    path_fields = ['PeiGuMingXiList']

    def get_original_time_field(self):
        return 'PeiGuGongGaoRi'

    def get_data_map(self):
        return {
            "rights_issues": ("ShiJiPeiGu", to_float),
            "rights_issue_price": ("PeiGuJiaGe", to_float),
            "rights_raising_fund": ("ShiJiMuJi", to_float)
        }

    def on_finish(self):
        last_year = str(now_pd_timestamp().year)
        codes = [item.code for item in self.entities]
        need_filleds = DividendFinancing.query_data(provider=self.provider, codes=codes,
                                                    return_type='domain',
                                                    session=self.session,
                                                    filters=[DividendFinancing.rights_raising_fund.is_(None)],
                                                    end_timestamp=last_year)

        for item in need_filleds:
            df = RightsIssueDetail.query_data(provider=self.provider, entity_id=item.entity_id,
                                              columns=[RightsIssueDetail.timestamp,
                                                       RightsIssueDetail.rights_raising_fund],
                                              start_timestamp=item.timestamp,
                                              end_timestamp="{}-12-31".format(item.timestamp.year))
            if pd_is_not_null(df):
                item.rights_raising_fund = df['rights_raising_fund'].sum()
                self.session.commit()

        super().on_finish()


if __name__ == '__main__':
    # init_log('rights_issue.log')

    recorder = RightsIssueDetailRecorder(codes=SAMPLE_STOCK_CODES)
    recorder.run()
# the __all__ is generated
__all__ = ['RightsIssueDetailRecorder']


# -*- coding: utf-8 -*-
import json
from typing import Type

from zvtm.contract.api import del_data, get_db_session
from zvtm.contract.schema import Mixin


class StatefulService(object):
    state_schema: Type[Mixin] = None
    name = None

    def __init__(self) -> None:
        assert self.state_schema is not None
        self.state_session = get_db_session(data_schema=self.state_schema, provider='zvt')
        self.state_domain = self.state_schema.get_one(id=self.get_state_entity_id())
        if self.state_domain:
            self.state: dict = self.decode_state(self.state_domain.state)
        else:
            self.state = None

    def get_state_entity_id(self):
        if self.name is not None:
            return self.name
        return type(self).__name__.lower()

    def clear_state_data(self):
        del_data(self.state_schema, filters=[self.state_schema.entity_id == self.get_state_entity_id()])

    def decode_state(self, state: str):
        # 反序列化
        return json.loads(state, object_hook=self.state_object_hook())

    def encode_state(self, state: object):
        # 序列化
        return json.dumps(state, cls=self.state_encoder())

    def state_object_hook(self):
        return None

    def state_encoder(self):
        return None

    def persist_state(self):
        state_str = self.encode_state(self.state)
        if not self.state_domain:
            self.state_domain = self.state_schema(id=self.get_state_entity_id(), entity_id=self.get_state_entity_id())
        self.state_domain.state = state_str
        self.state_session.add(self.state_domain)
        self.state_session.commit()


# the __all__ is generated
__all__ = ['StatefulService']



# -*- coding: utf-8 -*-
from sqlalchemy import Column, Float, String
from sqlalchemy.orm import declarative_base

from zvtm.contract import Mixin
from zvtm.contract.register import register_schema

Stock1dMaFactorBase = declarative_base()


class Stock1dMaFactor(Stock1dMaFactorBase, Mixin):
    __tablename__ = 'Stock1dMaFactor'

    level = Column(String(length=32))
    code = Column(String(length=32))
    name = Column(String(length=32))

    open = Column(Float)
    close = Column(Float)
    high = Column(Float)
    low = Column(Float)

    ma5 = Column(Float)
    ma10 = Column(Float)

    ma34 = Column(Float)
    ma55 = Column(Float)
    ma89 = Column(Float)
    ma144 = Column(Float)

    ma120 = Column(Float)
    ma250 = Column(Float)


register_schema(providers=['zvt'], db_name='stock_1d_ma_factor', schema_base=Stock1dMaFactorBase)
# the __all__ is generated
__all__ = ['Stock1dMaFactor']


# -*- coding: utf-8 -*-
def to_string(cls):
    def __str__(self):
        return '%s(%s)' % (
            type(self).__name__,
            ', '.join('%s=%s' % item for item in vars(self).items())
        )

    cls.__str__ = __str__
    return cls
# the __all__ is generated
__all__ = ['to_string']


# -*- coding: utf-8 -*-

import pandas as pd
from jqdatapy.api import get_fundamentals
from pandas._libs.tslibs.timedeltas import Timedelta

from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import TimeSeriesDataRecorder
from zvtm.domain import Stock, StockValuation, Etf,StockDetail
from zvtm.recorders.joinquant.common import to_jq_entity_id
from zvtm.utils.time_utils import now_pd_timestamp, to_time_str, to_pd_timestamp


class JqChinaStockValuationRecorder(TimeSeriesDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Stock

    # 数据来自jq
    provider = 'joinquant'

    data_schema = StockValuation

    def init_entities(self):
        super().init_entities()
        # 过滤掉退市的
        self.entities = [entity for entity in self.entities if
                         (entity.end_date is None) or (entity.end_date > now_pd_timestamp())]

    def record(self, entity, start, end, size, timestamps):
        start = max(start, to_pd_timestamp('2005-01-01'))
        end = min(now_pd_timestamp(), start + Timedelta(days=500))

        count: Timedelta = end - start
        #增加对上下市日期判断
        # dfstockdetail = StockDetail.query_data(provider='joinquant',code=entity.code, return_type='df')
        # if start > dfstockdetail['end_date'][0]:
        #     return;
        # df = get_fundamentals_continuously(q, end_date=now_time_str(), count=count.days + 1, panel=False)
        df = get_fundamentals(table='valuation', code=to_jq_entity_id(entity), date=to_time_str(end),
                              count=min(count.days, 500))
        df['entity_id'] = entity.id
        df['timestamp'] = pd.to_datetime(df['day'])
        df['code'] = entity.code
        df['name'] = entity.name
        df['id'] = df['timestamp'].apply(lambda x: "{}_{}".format(entity.id, to_time_str(x)))
        df = df.rename({'pe_ratio_lyr': 'pe',
                        'pe_ratio': 'pe_ttm',
                        'pb_ratio': 'pb',
                        'ps_ratio': 'ps',
                        'pcf_ratio': 'pcf'},
                       axis='columns')

        df['market_cap'] = df['market_cap'] * 100000000
        df['circulating_market_cap'] = df['circulating_market_cap'] * 100000000
        df['capitalization'] = df['capitalization'] * 10000
        df['circulating_cap'] = df['circulating_cap'] * 10000
        df['turnover_ratio'] = df['turnover_ratio'] * 0.01
        df_to_db(df=df, data_schema=self.data_schema, provider=self.provider, force_update=self.force_update)

        return None


if __name__ == '__main__':
    # 上证50
    df = Etf.get_stocks(code='510050')
    stocks = df.stock_id.tolist()
    print(stocks)
    print(len(stocks))

    JqChinaStockValuationRecorder(entity_ids=['stock_sz_300999'], force_update=True).run()
# the __all__ is generated
__all__ = ['JqChinaStockValuationRecorder']


# -*- coding: utf-8 -*-
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule jq_margin_trading_recorder
from .jq_margin_trading_recorder import *
from .jq_margin_trading_recorder import __all__ as _jq_margin_trading_recorder_all
__all__ += _jq_margin_trading_recorder_all

# import all from submodule jq_stock_valuation_recorder
from .jq_stock_valuation_recorder import *
from .jq_stock_valuation_recorder import __all__ as _jq_stock_valuation_recorder_all
__all__ += _jq_stock_valuation_recorder_all

# import all from submodule jq_etf_valuation_recorder
from .jq_etf_valuation_recorder import *
from .jq_etf_valuation_recorder import __all__ as _jq_etf_valuation_recorder_all
__all__ += _jq_etf_valuation_recorder_all


# -*- coding: utf-8 -*-

import datetime
import os
import zipfile


def zip_dir(src_dir,
            dst_dir=None,
            zip_file_name=None):
    if not zip_file_name:
        zip_file_name = "data-{}.zip".format(datetime.datetime.today())

    if dst_dir:
        dst_path = os.path.join(dst_dir, zip_file_name)
    else:
        dst_path = zip_file_name

    # os.remove(dst_path)

    the_zip_file = zipfile.ZipFile(dst_path, 'w')

    for folder, subfolders, files in os.walk(src_dir):
        for file in files:
            the_path = os.path.join(folder, file)
            # if 'zvt_business.db' in the_path:
            #     continue
            print("zip {}".format(the_path))
            the_zip_file.write(the_path,
                               os.path.relpath(the_path, src_dir),
                               compress_type=zipfile.ZIP_DEFLATED)

    the_zip_file.close()


def unzip(zip_file, dst_dir):
    the_zip_file = zipfile.ZipFile(zip_file)
    print("start unzip {} to {}".format(zip_file, dst_dir))
    the_zip_file.extractall(dst_dir)
    print("finish unzip {} to {}".format(zip_file, dst_dir))
    the_zip_file.close()
# the __all__ is generated
__all__ = ['zip_dir', 'unzip']


# -*- coding: utf-8 -*-

from typing import List

import pandas as pd


from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import TimestampsDataRecorder
from zvtm.domain import Index, IndexStock
#from zvtm.recorders.exchange.api import cs_index_stock_api, cn_index_stock_api

from zvtm.utils.time_utils import pre_month_start_date

from jqdatasdk import get_index_stocks,auth,get_index_weights
from zvtm import zvt_config
from jqdatapy.api import get_all_securities, run_query
from zvtm.contract.recorder import Recorder
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.domain import EtfStock, Stock, Etf, StockDetail,Index,IndexStock
from zvtm.utils.time_utils import to_pd_timestamp, to_time_str, TIME_FORMAT_MON,TIME_FORMAT_DAY,to_timestamp
from zvtm.recorders.joinquant.common import to_entity_id
from zvtm.contract.api import df_to_db, get_entity_exchange, get_entity_code


class JqIndexRecorder(Recorder):# get_index(code, self):
    provider = 'exchange'
    data_schema = Index

    def to_zvt_entity(self, df, entity_type, category=None):
        df = df.set_index('code')
        df.index.name = 'entity_id'
        df = df.reset_index()
        # 上市日期
        df.rename(columns={'start_date': 'timestamp'}, inplace=True)
        df['timestamp'] = pd.to_datetime(df['timestamp'])
        df['list_date'] = df['timestamp']
        df['end_date'] = pd.to_datetime(df['end_date'])

        df['entity_id'] = df['entity_id'].apply(lambda x: to_entity_id(entity_type=entity_type, jq_code=x))
        df['id'] = df['entity_id']
        df['entity_type'] = entity_type
        df['exchange'] = df['entity_id'].apply(lambda x: get_entity_exchange(x))
        df['code'] = df['entity_id'].apply(lambda x: get_entity_code(x))
        df['name'] = df['display_name']

        if category:
            df['category'] = category

        return df

    def run(self):
        # 抓取指数列表
        df_index = self.to_zvt_entity(get_all_securities(code='index'), entity_type='index', category='index')
        df_to_db(df_index, data_schema=Index, provider=self.provider, force_update=self.force_update)
        self.logger.info(f"finish record index:jqindex")


class JqIndexStockRecorder(TimestampsDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Index

    provider = 'joinquant'
    data_schema = IndexStock

    def __init__(self,
                 force_update=False,
                 sleeping_time=0,
                 exchanges=None,
                 entity_ids=None,
                 code=None,
                 codes=None,
                 day_data=False,
                 entity_filters=None,
                 ignore_failed=True,
                 real_time=False,
                 fix_duplicate_way='add',
                 start_timestamp=None,
                 end_timestamp=None,
                 record_history=False) -> None:
        self.code = code
        # super().__init__(force_update, sleeping_time, exchanges, entity_ids, code, codes, day_data, entity_filters,
        #                  ignore_failed, real_time, fix_duplicate_way, start_timestamp, end_timestamp)
        # self.record_history = record_history

    # def init_timestamps(self, entity_item) -> List[pd.Timestamp]:
    #     last_valid_date = pre_month_start_date()
    #     if self.record_history:
    #         # 每个月记录一次
    #         return [to_pd_timestamp(item) for item in
    #                 pd.date_range(entity_item.list_date, last_valid_date, freq='M')]
    #     else:
    #         return [last_valid_date]
    #start, end, size,  timestamps
    #for timestamp in timestamps:
    #不用init了，直接填充entity timestamp

    def run(self):#, entity,timestamp
        entity = Index.query_data(code=self.code)
        if entity.shape[0]==0:
            return
        timestamp = pd.Timestamp.now()
        df = get_index_stock(entity_item=entity, timestamp=timestamp, name=entity.name)#
        df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                     force_update=True)#self.force_update


def get_index_stock(entity_item, timestamp, name):
    entity_type = 'index'
    jqstock = entity_item.code[0] + '.XSHG'
    auth(zvt_config['jq_username'], zvt_config['jq_password'])
    data_str = to_time_str(timestamp, TIME_FORMAT_DAY)
    #根据index指数的code编码获取包含的股票series
    df_index_stock = get_index_stocks(jqstock,date=data_str)
    df_index_stock_weight = get_index_weights(index_id=jqstock)
    the_list = []
    if len(df_index_stock) > 0:
        for istock in df_index_stock:
            result = Stock.query_data(provider='joinquant', code=istock[0:6])

            if result.shape[0] == 0:
                return
            stock_code = istock
            stock_id = result['entity_id'][0]
            exchange = result['exchange'][0]
            entity_id = f'{entity_item.entity_id[0]}_{stock_id}'
            #f'{entity_type}_{exchange}_{istock}'

            stock_name = result['name'][0]
            code = entity_item.code[0]
            name = entity_item.name[0]
            # 指数股票权重
            if df_index_stock_weight.shape[0]>0:
                report_date = df_index_stock_weight['date'][0]
                proportion = df_index_stock_weight['weight'][0]
            else:
                report_date = ''
                proportion =''
            #report_period = ''
            #shares = ''
            #'market_cap': '',  # Market capitalization/capitalisation 市场总值
            the_list.append({
                'id': '{}_{}_{}'.format(entity_item.entity_id[0], str(to_timestamp(timestamp))[:8], stock_id),
                'entity_id': entity_id,
                'entity_type': entity_type,
                'exchange': exchange,
                'code': code,
                'name': name,
                'timestamp': report_date,
                'stock_id': stock_id,
                'stock_code': stock_code[:6],
                'stock_name': stock_name,
                'proportion': proportion,
                'report_date':report_date
            })
        if the_list:
            df = pd.DataFrame.from_records(the_list)
            return df












if __name__ == '__main__':
    #JqIndexRecorder().run()
    JqIndexStockRecorder(code=['000001']).run()
# the __all__ is generated
__all__ = ['JqIndexRecorder','JqIndexStockRecorder']


# -*- coding: utf-8 -*-
from zvtm.domain import DividendDetail
from zvtm.recorders.eastmoney.common import EastmoneyPageabeDataRecorder
from zvtm.utils.time_utils import to_pd_timestamp


class DividendDetailRecorder(EastmoneyPageabeDataRecorder):
    data_schema = DividendDetail

    url = 'https://emh5.eastmoney.com/api/FenHongRongZi/GetFenHongSongZhuanList'
    page_url = url
    path_fields = ['FenHongSongZhuanList']

    def get_original_time_field(self):
        return 'GongGaoRiQi'

    def get_data_map(self):
        return {
            # 公告日
            "announce_date": ("GongGaoRiQi", to_pd_timestamp),
            # 股权登记日
            "record_date": ("GuQuanDengJiRi", to_pd_timestamp),
            # 除权除息日
            "dividend_date": ("ChuQuanChuXiRi", to_pd_timestamp),
            # 方案
            "dividend": ("FengHongFangAn", str)
        }


if __name__ == '__main__':
    # init_log('dividend_detail.log')

    recorder = DividendDetailRecorder(codes=['601318'])
    recorder.run()
# the __all__ is generated
__all__ = ['DividendDetailRecorder']


# -*- coding: utf-8 -*-
import argparse

import pandas as pd
from jqdatapy.api import get_token, get_bars

from zvtm import init_log, zvt_config
from zvtm.api.kdata import generate_kdata_id, get_kdata_schema, get_kdata
from zvtm.contract import IntervalLevel
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import FixedCycleDataRecorder
from zvtm.domain import Index, IndexKdataCommon
from zvtm.recorders.joinquant.common import to_jq_trading_level, to_jq_entity_id
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.utils.time_utils import to_time_str, TIME_FORMAT_DAY, TIME_FORMAT_ISO8601


class JqChinaIndexKdataRecorder(FixedCycleDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Index

    # 数据来自jq
    provider = 'joinquant'

    # 只是为了把recorder注册到data_schema
    data_schema = IndexKdataCommon

    def __init__(self,
                 force_update=True,
                 sleeping_time=10,
                 exchanges=None,
                 entity_ids=None,
                 code=None,
                 codes=None,
                 day_data=False,
                 entity_filters=None,
                 ignore_failed=True,
                 real_time=False,
                 fix_duplicate_way='ignore',
                 start_timestamp=None,
                 end_timestamp=None,
                 level=IntervalLevel.LEVEL_1DAY,
                 kdata_use_begin_time=False,
                 one_day_trading_minutes=24 * 60) -> None:
        level = IntervalLevel(level)
        self.data_schema = get_kdata_schema(entity_type='index', level=level)
        self.jq_trading_level = to_jq_trading_level(level)
        get_token(zvt_config['jq_username'], zvt_config['jq_password'], force=True)
        super().__init__(force_update, sleeping_time, exchanges, entity_ids, code, codes, day_data, entity_filters,
                         ignore_failed, real_time, fix_duplicate_way, start_timestamp, end_timestamp, level,
                         kdata_use_begin_time, one_day_trading_minutes)

    def init_entities(self):
        super().init_entities()
        # ignore no data index
        self.entities = [entity for entity in self.entities if
                         entity.code not in ['310001', '310002', '310003', '310004']]

    def generate_domain_id(self, entity, original_data):
        return generate_kdata_id(entity_id=entity.id, timestamp=original_data['timestamp'], level=self.level)

    def record(self, entity, start, end, size, timestamps):
        if not self.end_timestamp:
            df = get_bars(to_jq_entity_id(entity),
                          count=size,
                          unit=self.jq_trading_level,
                          # fields=['date', 'open', 'close', 'low', 'high', 'volume', 'money']
                          )
        else:
            end_timestamp = to_time_str(self.end_timestamp)
            df = get_bars(to_jq_entity_id(entity),
                          count=size,
                          unit=self.jq_trading_level,
                          # fields=['date', 'open', 'close', 'low', 'high', 'volume', 'money'],
                          end_date=end_timestamp)
        if pd_is_not_null(df):
            df['name'] = entity.name
            df.rename(columns={'money': 'turnover', 'date': 'timestamp'}, inplace=True)

            df['entity_id'] = entity.id
            df['timestamp'] = pd.to_datetime(df['timestamp'])
            df['provider'] = 'joinquant'
            df['level'] = self.level.value
            df['code'] = entity.code

            def generate_kdata_id(se):
                if self.level >= IntervalLevel.LEVEL_1DAY:
                    return "{}_{}".format(se['entity_id'], to_time_str(se['timestamp'], fmt=TIME_FORMAT_DAY))
                else:
                    return "{}_{}".format(se['entity_id'], to_time_str(se['timestamp'], fmt=TIME_FORMAT_ISO8601))

            df['id'] = df[['entity_id', 'timestamp']].apply(generate_kdata_id, axis=1)

            df = df.drop_duplicates(subset='id', keep='last')

            df_to_db(df=df, data_schema=self.data_schema, provider=self.provider, force_update=self.force_update)

        return None


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--level', help='trading level', default='1mon', choices=[item.value for item in IntervalLevel])# 1d 1wk 1mon
    parser.add_argument('--codes', help='codes', default=['000001'], nargs='+')

    args = parser.parse_args()

    level = IntervalLevel(args.level)
    codes = args.codes

    init_log('jq_china_stock_{}_kdata.log'.format(args.level))
    JqChinaIndexKdataRecorder(level=level, sleeping_time=0, codes=codes, real_time=False).run()

    print(get_kdata(entity_id='index_sh_000001', limit=10))
# the __all__ is generated
__all__ = ['JqChinaIndexKdataRecorder']


# -*- coding: utf-8 -*-

from typing import List

from zvtm.api.utils import get_recent_report_date
from zvtm.contract import PortfolioStockHistory
from zvtm.contract.api import get_schema_by_name
from zvtm.domain import ReportPeriod, Fund, Etf
from zvtm.utils.time_utils import to_pd_timestamp, now_pd_timestamp


def portfolio_relate_stock(df, portfolio):
    df['entity_id'] = portfolio.entity_id
    df['entity_type'] = portfolio.entity_type
    df['exchange'] = portfolio.exchange
    df['code'] = portfolio.code
    df['name'] = portfolio.name

    return df


# 季报只有前十大持仓，半年报和年报才有全量的持仓信息，故根据离timestamp最近的报表(年报 or 半年报)来确定持仓
def get_portfolio_stocks(portfolio_entity=Fund, code=None, codes=None, ids=None, timestamp=now_pd_timestamp(),
                         provider=None):
    portfolio_stock = f'{portfolio_entity.__name__}Stock'
    data_schema: PortfolioStockHistory = get_schema_by_name(portfolio_stock)
    latests: List[PortfolioStockHistory] = data_schema.query_data(provider=provider, code=code, end_timestamp=timestamp,
                                                                  order=data_schema.timestamp.desc(), limit=1,
                                                                  return_type='domain')
    if latests:
        latest_record = latests[0]
        # 获取最新的报表
        df = data_schema.query_data(provider=provider, code=code, codes=codes, ids=ids, end_timestamp=timestamp,
                                    filters=[data_schema.report_date == latest_record.report_date])
        # 最新的为年报或者半年报
        if latest_record.report_period == ReportPeriod.year or latest_record.report_period == ReportPeriod.half_year:
            return df
        # 季报，需要结合 年报或半年报 来算持仓
        else:
            step = 0
            while step <= 20:
                report_date = get_recent_report_date(latest_record.report_date, step=step)

                pre_df = data_schema.query_data(provider=provider, code=code, codes=codes, ids=ids,
                                                end_timestamp=timestamp,
                                                filters=[data_schema.report_date == to_pd_timestamp(report_date)])
                df = df.append(pre_df)

                # 半年报和年报
                if (ReportPeriod.half_year.value in pre_df['report_period'].tolist()) or (
                        ReportPeriod.year.value in pre_df['report_period'].tolist()):
                    # 保留最新的持仓
                    df = df.drop_duplicates(subset=['stock_code'], keep='first')
                    return df
                step = step + 1


def get_etf_stocks(code=None, codes=None, ids=None, timestamp=now_pd_timestamp(), provider=None):
    return get_portfolio_stocks(portfolio_entity=Etf, code=code, codes=codes, ids=ids, timestamp=timestamp,
                                provider=provider)


def get_fund_stocks(code=None, codes=None, ids=None, timestamp=now_pd_timestamp(), provider=None):
    return get_portfolio_stocks(portfolio_entity=Fund, code=code, codes=codes, ids=ids, timestamp=timestamp,
                                provider=provider)
# the __all__ is generated
__all__ = ['portfolio_relate_stock', 'get_portfolio_stocks', 'get_etf_stocks', 'get_fund_stocks']


# -*- coding: utf-8 -*-
import io
import logging

import pandas as pd
import requests

from zvtm.api import china_stock_code_to_id
from zvtm.recorders.consts import DEFAULT_HEADER
from zvtm.utils import now_pd_timestamp

logger = logging.getLogger(__name__)

original_page_url = 'http://www.csindex.com.cn/zh-CN/downloads/indices'
url = 'http://www.csindex.com.cn/uploads/file/autofile/cons/{}cons.xls'


def get_cs_index_stock(code, timestamp, name=None):
    entity_type = 'index'
    exchange = 'sh'
    entity_id = f'{entity_type}_{exchange}_{code}'

    response = requests.get(url.format(code), headers=DEFAULT_HEADER)
    response.raise_for_status()

    df = pd.read_excel(io.BytesIO(response.content))

    df = df[['日期Date', '成分券代码Constituent Code', '成分券名称Constituent Name']].rename(
        columns={'日期Date': 'timestamp', '成分券代码Constituent Code': 'stock_code',
                 '成分券名称Constituent Name': 'stock_name'})

    df['entity_id'] = entity_id
    df['entity_type'] = 'index'
    df['exchange'] = 'sh'
    df['code'] = code
    df['name'] = name
    df['stock_id'] = df['stock_code'].apply(lambda x: china_stock_code_to_id(str(x)))
    # id format: {entity_id}_{timestamp}_{stock_id}
    df['id'] = df[['entity_id', 'timestamp', 'stock_id']].apply(lambda x: '_'.join(x.astype(str)), axis=1)
    df['timestamp'] = pd.to_datetime(df['timestamp'])

    return df


if __name__ == '__main__':
    df = get_cs_index_stock(code='000001', name='上证指数', timestamp=now_pd_timestamp())
    print(df)
# the __all__ is generated
__all__ = ['get_cs_index_stock']



# -*- coding: utf-8 -*-
from enum import Enum


class IntervalLevel(Enum):
    LEVEL_TICK = 'tick'
    LEVEL_1MIN = '1m'
    LEVEL_5MIN = '5m'
    LEVEL_15MIN = '15m'
    LEVEL_30MIN = '30m'
    LEVEL_1HOUR = '1h'
    LEVEL_4HOUR = '4h'
    LEVEL_1DAY = '1d'
    LEVEL_1WEEK = '1wk'
    LEVEL_1MON = '1mon'

    def to_pd_freq(self):
        if self == IntervalLevel.LEVEL_1MIN:
            return '1min'
        if self == IntervalLevel.LEVEL_5MIN:
            return '5min'
        if self == IntervalLevel.LEVEL_15MIN:
            return '15min'
        if self == IntervalLevel.LEVEL_30MIN:
            return '30min'
        if self == IntervalLevel.LEVEL_1HOUR:
            return '1H'
        if self == IntervalLevel.LEVEL_4HOUR:
            return '4H'
        if self >= IntervalLevel.LEVEL_1DAY:
            return '1D'

    def floor_timestamp(self, pd_timestamp):
        if self == IntervalLevel.LEVEL_1MIN:
            return pd_timestamp.floor('1min')
        if self == IntervalLevel.LEVEL_5MIN:
            return pd_timestamp.floor('5min')
        if self == IntervalLevel.LEVEL_15MIN:
            return pd_timestamp.floor('15min')
        if self == IntervalLevel.LEVEL_30MIN:
            return pd_timestamp.floor('30min')
        if self == IntervalLevel.LEVEL_1HOUR:
            return pd_timestamp.floor('1h')
        if self == IntervalLevel.LEVEL_4HOUR:
            return pd_timestamp.floor('4h')
        if self == IntervalLevel.LEVEL_1DAY:
            return pd_timestamp.floor('1d')

    def to_minute(self):
        return int(self.to_second() / 60)

    def to_second(self):
        return int(self.to_ms() / 1000)

    def to_ms(self):
        # we treat tick intervals is 5s, you could change it
        if self == IntervalLevel.LEVEL_TICK:
            return 5 * 1000
        if self == IntervalLevel.LEVEL_1MIN:
            return 60 * 1000
        if self == IntervalLevel.LEVEL_5MIN:
            return 5 * 60 * 1000
        if self == IntervalLevel.LEVEL_15MIN:
            return 15 * 60 * 1000
        if self == IntervalLevel.LEVEL_30MIN:
            return 30 * 60 * 1000
        if self == IntervalLevel.LEVEL_1HOUR:
            return 60 * 60 * 1000
        if self == IntervalLevel.LEVEL_4HOUR:
            return 4 * 60 * 60 * 1000
        if self == IntervalLevel.LEVEL_1DAY:
            return 24 * 60 * 60 * 1000
        if self == IntervalLevel.LEVEL_1WEEK:
            return 7 * 24 * 60 * 60 * 1000
        if self == IntervalLevel.LEVEL_1MON:
            return 31 * 7 * 24 * 60 * 60 * 1000

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.to_ms() >= other.to_ms()
        return NotImplemented

    def __gt__(self, other):

        if self.__class__ is other.__class__:
            return self.to_ms() > other.to_ms()
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.to_ms() <= other.to_ms()
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.to_ms() < other.to_ms()
        return NotImplemented


class AdjustType(Enum):
    # 这里用拼音，因为英文不直观 split-adjusted？wtf?
    # 不复权
    bfq = 'bfq'
    # 前复权
    qfq = 'qfq'
    # 后复权
    hfq = 'hfq'


class ActorType(Enum):
    # 个人
    individual = 'individual'
    # 公募基金
    raised_fund = 'raised_fund'
    # 社保
    social_security = 'social_security'
    # 保险
    insurance = 'insurance'
    # 外资
    qfii = 'qfii'
    # 信托
    trust = 'trust'
    # 券商
    broker = 'broker'
    # 私募
    private_equity = 'private_equity'
    # 公司(可能包括私募)
    corporation = 'corporation'


class TradableType(Enum):
    # A股(中国)
    stock = 'stock'
    # 美股
    stockus = 'stockus'
    # 港股
    stockhk = 'stockhk'
    # 期货(中国)
    future = 'future'
    # 数字货币
    coin = 'coin'
    # 期权
    option = 'option'
    # 基金
    fund = 'fund'


class Exchange(Enum):
    # 上证交易所
    sh = 'sh'
    # 深证交易所
    sz = 'sz'

    # 对于中国的非交易所的 标的
    cn = 'cn'

    # 纳斯达克
    nasdaq = 'nasdaq'

    # 纽交所
    nyse = 'nyse'

    # 港交所
    hk = 'hk'

    # 数字货币
    binance = 'binance'
    huobipro = 'huobipro'

    # 上海期货交易所
    shfe = 'shfe'
    # 大连商品交易所
    dce = "dce"
    # 郑州商品交易所
    czce = 'czce'
    # 中国金融期货交易所
    cffex = 'cffex'


tradable_type_map_exchanges = {
    TradableType.stock: [Exchange.sh, Exchange.sz],
    TradableType.stockhk: [Exchange.hk],
    TradableType.stockus: [Exchange.nasdaq, Exchange.nyse],

    TradableType.future: [Exchange.shfe, Exchange.dce, Exchange.czce, Exchange.cffex],

    TradableType.coin: [Exchange.binance, Exchange.huobipro]
}


def get_entity_exchanges(entity_type):
    entity_type = TradableType(entity_type)
    return tradable_type_map_exchanges.get(entity_type)


from . import zvt_context

# the __all__ is generated
__all__ = ['IntervalLevel', 'AdjustType', 'ActorType', 'TradableType', 'Exchange', 'zvt_context']

# import all from submodule schema
from .schema import *
from .schema import __all__ as _schema_all

__all__ += _schema_all



# -*- coding: utf-8 -*-
from sqlalchemy import Column, String, Float
from sqlalchemy.orm import declarative_base

from zvtm.contract import Mixin
from zvtm.contract.register import register_schema

TradingBase = declarative_base()


class ManagerTrading(TradingBase, Mixin):
    __tablename__ = 'manager_trading'

    provider = Column(String(length=32))
    code = Column(String(length=32))
    # 日期 变动人 变动数量(股) 交易均价(元) 结存股票(股) 交易方式 董监高管 高管职位 与高管关系
    # 2017-08-11 韦春 200 9.16 -- 竞价交易 刘韬 高管 兄弟姐妹

    # 变动人
    trading_person = Column(String(length=32))
    # 变动数量
    volume = Column(Float)
    # 交易均价
    price = Column(Float)
    # 结存股票
    holding = Column(Float)
    # 交易方式
    trading_way = Column(String(length=32))
    # 董监高管
    manager = Column(String(length=32))
    # 高管职位
    manager_position = Column(String(length=32))
    # 与高管关系
    relationship_with_manager = Column(String(length=32))


class HolderTrading(TradingBase, Mixin):
    __tablename__ = 'holder_trading'

    provider = Column(String(length=32))
    code = Column(String(length=32))

    # 股东名称
    holder_name = Column(String(length=128))
    # 变动数量
    volume = Column(Float)
    # 变动比例
    change_pct = Column(Float)
    # 变动后持股比例
    holding_pct = Column(Float)


class BigDealTrading(TradingBase, Mixin):
    __tablename__ = 'big_deal_trading'

    provider = Column(String(length=32))
    code = Column(String(length=32))

    # 成交额
    turnover = Column(Float)
    # 成交价
    price = Column(Float)
    # 卖出营业部
    sell_broker = Column(String(length=128))
    # 买入营业部
    buy_broker = Column(String(length=128))
    # 折/溢价率
    compare_rate = Column(Float)


class MarginTrading(TradingBase, Mixin):
    __tablename__ = 'margin_trading'
    code = Column(String(length=32))

    # 融资余额(元）
    fin_value = Column(Float)
    # 融资买入额（元）
    fin_buy_value = Column(Float)
    # 融资偿还额（元）
    fin_refund_value = Column(Float)
    # 融券余量（股）
    sec_value = Column(Float)
    # 融券卖出量（股）
    sec_sell_value = Column(Float)
    # 融券偿还量（股）
    sec_refund_value = Column(Float)
    # 融资融券余额（元）
    fin_sec_value = Column(Float)


class DragonAndTiger(TradingBase, Mixin):
    __tablename__ = 'dragon_and_tiger'

    provider = Column(String(length=32))
    code = Column(String(length=32))

    # 异动原因
    reason = Column(String(length=128))
    # 成交额
    turnover = Column(Float)
    # 涨幅
    change_pct = Column(Float)
    # 净买入
    net_inflow = Column(Float)

    # 买入营业部
    net_in_dep1 = Column(String(length=128))
    net_in_dep1_money_in = Column(Float)
    net_in_dep1_money_out = Column(Float)
    net_in_dep1_rate = Column(Float)

    net_in_dep2 = Column(String(length=128))
    net_in_dep2_money_in = Column(Float)
    net_in_dep2_money_out = Column(Float)
    net_in_dep2_rate = Column(Float)

    net_in_dep3 = Column(String(length=128))
    net_in_dep3_money_in = Column(Float)
    net_in_dep3_money_out = Column(Float)
    net_in_dep3_rate = Column(Float)

    net_in_dep4 = Column(String(length=128))
    net_in_dep4_money_in = Column(Float)
    net_in_dep4_money_out = Column(Float)
    net_in_dep4_rate = Column(Float)

    net_in_dep5 = Column(String(length=128))
    net_in_dep5_money_in = Column(Float)
    net_in_dep5_money_out = Column(Float)
    net_in_dep5_rate = Column(Float)

    # 卖出营业部
    net_out_dep1 = Column(String(length=128))
    net_out_dep1_money_in = Column(Float)
    net_out_dep1_money_out = Column(Float)
    net_out_dep1_rate = Column(Float)

    net_out_dep2 = Column(String(length=128))
    net_out_dep2_money_in = Column(Float)
    net_out_dep2_money_out = Column(Float)
    net_out_dep2_rate = Column(Float)

    net_out_dep3 = Column(String(length=128))
    net_out_dep3_money_in = Column(Float)
    net_out_dep3_money_out = Column(Float)
    net_out_dep3_rate = Column(Float)

    net_out_dep4 = Column(String(length=128))
    net_out_dep4_money_in = Column(Float)
    net_out_dep4_money_out = Column(Float)
    net_out_dep4_rate = Column(Float)

    net_out_dep5 = Column(String(length=128))
    net_out_dep5_money_in = Column(Float)
    net_out_dep5_money_out = Column(Float)
    net_out_dep5_rate = Column(Float)


register_schema(providers=['eastmoney', 'joinquant'], db_name='trading', schema_base=TradingBase, entity_type='stock')

# the __all__ is generated
__all__ = ['ManagerTrading', 'HolderTrading', 'BigDealTrading', 'MarginTrading', 'DragonAndTiger']


# -*- coding: utf-8 -*-#
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule generator
from .generator import *
from .generator import __all__ as _generator_all
__all__ += _generator_all

# import all from submodule templates
from .templates import *
from .templates import __all__ as _templates_all
__all__ += _templates_all


# -*- coding: utf-8 -*-
import importlib
import json
import logging
import os
import pkgutil
import pprint
from logging.handlers import RotatingFileHandler

import pandas as pd
import pkg_resources
from pkg_resources import get_distribution, DistributionNotFound

from zvtm.consts import DATA_SAMPLE_ZIP_PATH, ZVT_TEST_HOME, ZVT_HOME, ZVT_TEST_DATA_PATH, ZVT_TEST_ZIP_DATA_PATH

try:
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = 'unknown'
finally:
    del get_distribution, DistributionNotFound

logger = logging.getLogger(__name__)


def init_log(file_name='zvtm.log', log_dir=None, simple_formatter=True):
    if not log_dir:
        log_dir = zvt_env['log_path']

    root_logger = logging.getLogger()

    # reset the handlers
    root_logger.handlers = []

    root_logger.setLevel(logging.INFO)

    file_name = os.path.join(log_dir, file_name)

    file_log_handler = RotatingFileHandler(file_name, maxBytes=524288000, backupCount=10)

    file_log_handler.setLevel(logging.INFO)

    console_log_handler = logging.StreamHandler()
    console_log_handler.setLevel(logging.INFO)

    # create formatter and add it to the handlers
    if simple_formatter:
        formatter = logging.Formatter(
            "%(asctime)s  %(levelname)s  %(threadName)s  %(message)s")
    else:
        formatter = logging.Formatter(
            "%(asctime)s  %(levelname)s  %(threadName)s  %(name)s:%(filename)s:%(lineno)s  %(funcName)s  %(message)s")
    file_log_handler.setFormatter(formatter)
    console_log_handler.setFormatter(formatter)

    # add the handlers to the logger
    root_logger.addHandler(file_log_handler)
    root_logger.addHandler(console_log_handler)

os.environ.setdefault('SQLALCHEMY_WARN_20',"1")
pd.set_option('expand_frame_repr', False)
pd.set_option('mode.chained_assignment', 'raise')
# pd.set_option('display.max_rows', None)
# pd.set_option('display.max_columns', None)

zvt_env = {}

zvt_config = {}

_plugins = {}


def init_env(zvt_home: str, **kwargs) -> dict:
    """
    init env

    :param zvt_home: home path for zvt
    """
    data_path = os.path.join(zvt_home, 'data')
    tmp_path = os.path.join(zvt_home, 'tmp')
    if not os.path.exists(data_path):
        os.makedirs(data_path)

    if not os.path.exists(tmp_path):
        os.makedirs(tmp_path)

    zvt_env['zvt_home'] = zvt_home
    zvt_env['data_path'] = data_path
    zvt_env['tmp_path'] = tmp_path

    # path for storing ui results
    zvt_env['ui_path'] = os.path.join(zvt_home, 'ui')
    if not os.path.exists(zvt_env['ui_path']):
        os.makedirs(zvt_env['ui_path'])

    # path for storing logs
    zvt_env['log_path'] = os.path.join(zvt_home, 'logs')
    if not os.path.exists(zvt_env['log_path']):
        os.makedirs(zvt_env['log_path'])

    init_log()

    pprint.pprint(zvt_env)

    # init config
    init_config(current_config=zvt_config, **kwargs)
    # init plugin
    # init_plugins()

    return zvt_env


def init_config(pkg_name: str = None, current_config: dict = None, **kwargs) -> dict:
    """
    init config
    """

    # create default config.json if not exist
    if pkg_name:
        config_file = f'{pkg_name}_config.json'
    else:
        pkg_name = 'zvt'
        config_file = 'config.json'

    logger.info(f'init config for {pkg_name}, current_config:{current_config}')

    config_path = os.path.join(zvt_env['zvt_home'], config_file)
    if not os.path.exists(config_path):
        from shutil import copyfile
        try:
            sample_config = pkg_resources.resource_filename(pkg_name, 'config.json')
            if os.path.exists(sample_config):
                copyfile(sample_config, config_path)
        except Exception as e:
            logger.warning(f'could not load config.json from package {pkg_name}')

    if os.path.exists(config_path):
        with open(config_path) as f:
            config_json = json.load(f)
            for k in config_json:
                current_config[k] = config_json[k]

    # set and save the config
    for k in kwargs:
        current_config[k] = kwargs[k]
        with open(config_path, 'w+') as outfile:
            json.dump(current_config, outfile)

    pprint.pprint(current_config)
    logger.info(f'current_config:{current_config}')

    return current_config


def init_plugins():
    for finder, name, ispkg in pkgutil.iter_modules():
        if name.startswith('zvt_'):
            try:
                _plugins[name] = importlib.import_module(name)
            except Exception as e:
                logger.warning(f'failed to load plugin {name}', e)
    logger.info(f'loaded plugins:{_plugins}')


if os.getenv('TESTING_ZVT'):
    init_env(zvt_home=ZVT_TEST_HOME)

    # init the sample data if need
    same = False
    if os.path.exists(ZVT_TEST_ZIP_DATA_PATH):
        import filecmp

        same = filecmp.cmp(ZVT_TEST_ZIP_DATA_PATH, DATA_SAMPLE_ZIP_PATH)

    if not same:
        from shutil import copyfile
        from zvtm.contract import *
        from zvtm.utils.zip_utils import unzip

        copyfile(DATA_SAMPLE_ZIP_PATH, ZVT_TEST_ZIP_DATA_PATH)
        unzip(ZVT_TEST_ZIP_DATA_PATH, ZVT_TEST_DATA_PATH)

else:
    init_env(zvt_home=ZVT_HOME)

# register to meta
import zvtm.contract as zvt_contract
import zvtm.recorders as zvt_recorders
import zvtm.factors as zvt_factors

__all__ = ['zvt_env', 'zvt_config', 'init_log', 'init_env', 'init_config', '__version__']



# -*- coding: utf-8 -*-
from typing import List, Union

import pandas as pd

from zvtm.contract import AdjustType
from zvtm.contract import IntervalLevel, TradableEntity
from zvtm.contract.drawer import Drawer
from zvtm.contract.factor import Accumulator
from zvtm.contract.factor import Transformer
from zvtm.contract.reader import DataReader
from zvtm.domain import Stock, Stock1dKdata
from zvtm.factors.technical_factor import TechnicalFactor
from zvtm.utils.time_utils import now_pd_timestamp


class MaUpTransformer(Transformer):
    def __init__(self, window=20) -> None:
        super().__init__()
        self.window = window

    def transform(self, input_df) -> pd.DataFrame:
        top_df = input_df['close'].groupby(level=0).rolling(window=self.window, min_periods=self.window).mean()
        top_df = top_df.reset_index(level=0, drop=True)
        input_df['mean'] = top_df


        return input_df


class MaUpFactor(TechnicalFactor):
    def __init__(self, entity_schema: TradableEntity = Stock, provider: str = None, entity_provider: str = None,
                 entity_ids: List[str] = None, exchanges: List[str] = None, codes: List[str] = None,
                 start_timestamp: Union[str, pd.Timestamp] = None, end_timestamp: Union[str, pd.Timestamp] = None,
                 columns: List = ['id', 'entity_id', 'timestamp', 'level', 'open', 'close', 'high', 'low'],
                 filters: List = None, order: object = None, limit: int = None,
                 level: Union[str, IntervalLevel] = IntervalLevel.LEVEL_1DAY, category_field: str = 'entity_id',
                 time_field: str = 'timestamp', computing_window: int = None, keep_all_timestamp: bool = False,
                 fill_method: str = 'ffill', effective_number: int = None,
                 accumulator: Accumulator = None, need_persist: bool = False, only_compute_factor: bool = False,
                 factor_name: str = None, clear_state: bool = False, only_load_factor: bool = False,
                 adjust_type: Union[AdjustType, str] = None, window=30) -> None:
        self.adjust_type = adjust_type

        transformer = MaUpTransformer(window=window)

        super().__init__(entity_schema, provider, entity_provider, entity_ids, exchanges, codes, start_timestamp,
                         end_timestamp, columns, filters, order, limit, level, category_field, time_field,
                         computing_window, keep_all_timestamp, fill_method, effective_number, transformer, accumulator,
                         need_persist, only_compute_factor, factor_name, clear_state, only_load_factor, adjust_type)


if __name__ == '__main__':
    factor = MaUpFactor(codes=['601099'], start_timestamp='2021-01-01',
                             end_timestamp=now_pd_timestamp(),
                             level=IntervalLevel.LEVEL_1DAY, window=60)
    print(factor.factor_df)

    Stock1dKdata.record_data(code='601099', provider='joinquant',start_timestamp='2021-01-01',
                             end_timestamp=now_pd_timestamp())
    data_reader1 = Stock1dKdata.query_data(code='601099', provider='joinquant',start_timestamp='2021-01-01',
                             end_timestamp=now_pd_timestamp())


    drawer = Drawer(main_df=data_reader1, factor_df_list=[factor.factor_df[['mean']]])
    drawer.draw_kline(show=True)
# the __all__ is generated
__all__ = ['MaUpTransformer', 'MaUpFactor']


# -*- coding: utf-8 -*-
import pandas as pd

from zvtm.contract import IntervalLevel
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import FixedCycleDataRecorder
from zvtm.domain import IndexMoneyFlow, Index, StockMoneyFlow
from zvtm.utils import pd_is_not_null, to_time_str


class JoinquantIndexMoneyFlowRecorder(FixedCycleDataRecorder):
    entity_provider = 'exchange'
    entity_schema = Index

    provider = 'joinquant'
    data_schema = IndexMoneyFlow

    def __init__(self, force_update=True, sleeping_time=10, exchanges=None, entity_ids=None, code=None,codes=None, day_data=False,
                 entity_filters=None, ignore_failed=True, real_time=False, fix_duplicate_way='ignore',
                 start_timestamp=None, end_timestamp=None, level=IntervalLevel.LEVEL_1DAY, kdata_use_begin_time=False,
                 one_day_trading_minutes=24 * 60) -> None:
        # 上证指数，深证成指，创业板指，科创板
        support_codes = ['000001', '399001', '399006', '000688']
        if not codes:
            codes = support_codes
        else:
            codes = list(set(codes) & set(support_codes))
        super().__init__(force_update, sleeping_time, exchanges, entity_ids, code, codes, day_data, entity_filters,
                         ignore_failed, real_time, fix_duplicate_way, start_timestamp, end_timestamp, level,
                         kdata_use_begin_time, one_day_trading_minutes)

    def record(self, entity, start, end, size, timestamps):
        # 上证
        if entity.code == '000001':
            all_df = StockMoneyFlow.query_data(provider=self.provider, start_timestamp=start,
                                               filters=[StockMoneyFlow.entity_id.like('stock_sh%')])
        # 深证
        elif entity.code == '399001':
            all_df = StockMoneyFlow.query_data(provider=self.provider, start_timestamp=start,
                                               filters=[StockMoneyFlow.entity_id.like('stock_sz%')])
        # 创业板
        elif entity.code == '399006':
            all_df = StockMoneyFlow.query_data(provider=self.provider, start_timestamp=start,
                                               filters=[StockMoneyFlow.code.like('300%')])
        # 科创板
        elif entity.code == '000688':
            all_df = StockMoneyFlow.query_data(provider=self.provider, start_timestamp=start,
                                               filters=[StockMoneyFlow.code.like('688%')])

        if pd_is_not_null(all_df):
            g = all_df.groupby('timestamp')
            for timestamp, df in g:
                se = pd.Series({'id': "{}_{}".format(entity.id, to_time_str(timestamp)),
                                'entity_id': entity.id,
                                'timestamp': timestamp,
                                'code': entity.code,
                                'name': entity.name})
                for col in ['net_main_inflows', 'net_huge_inflows', 'net_big_inflows', 'net_medium_inflows',
                            'net_small_inflows']:
                    se[col] = df[col].sum()

                for col in ['net_main_inflow_rate', 'net_huge_inflow_rate', 'net_big_inflow_rate',
                            'net_medium_inflow_rate',
                            'net_small_inflow_rate']:
                    se[col] = df[col].sum() / len(df)

                index_df = se.to_frame().T

                self.logger.info(index_df)

                df_to_db(df=index_df, data_schema=self.data_schema, provider=self.provider,
                         force_update=self.force_update)

        return None


if __name__ == '__main__':
    JoinquantIndexMoneyFlowRecorder(start_timestamp='2020-12-01').run()
# the __all__ is generated
__all__ = ['JoinquantIndexMoneyFlowRecorder']


# -*- coding: utf-8 -*-
from sqlalchemy import Column, String, DateTime, Boolean, Float, Integer, ForeignKey
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import relationship

from zvtm.contract import Mixin
from zvtm.contract.register import register_schema
from zvtm.utils import to_string

TraderBase = declarative_base()


# trader信息
class TraderInfo(TraderBase, Mixin):
    __tablename__ = 'trader_info'
    # 机器人名字
    trader_name = Column(String(length=128))

    entity_type = Column(String(length=128))
    start_timestamp = Column(DateTime)
    end_timestamp = Column(DateTime)
    provider = Column(String(length=32))
    level = Column(String(length=32))
    real_time = Column(Boolean)
    kdata_use_begin_time = Column(Boolean)
    kdata_adjust_type = Column(String(length=32))


# account stats of every day
@to_string
class AccountStats(TraderBase, Mixin):
    __tablename__ = 'account_stats'

    input_money = Column(Float)

    # 机器人名字
    trader_name = Column(String(length=128))
    # 可用现金
    cash = Column(Float)
    # 具体仓位
    positions = relationship("Position", back_populates="account_stats")
    # 市值
    value = Column(Float)
    # 市值+cash
    all_value = Column(Float)

    # 盈亏
    profit = Column(Float)
    # 盈亏比例
    profit_rate = Column(Float)

    # 收盘计算
    closing = Column(Boolean)


# the position for specific entity of every day
class Position(TraderBase, Mixin):
    __tablename__ = 'position'

    # 机器人名字
    trader_name = Column(String(length=128))
    # 账户id
    account_stats_id = Column(String(length=128), ForeignKey('account_stats.id'))
    account_stats = relationship("AccountStats", back_populates="positions")

    # 做多数量
    long_amount = Column(Float)
    # 可平多数量
    available_long = Column(Float)
    # 平均做多价格
    average_long_price = Column(Float)

    # 做空数量
    short_amount = Column(Float)
    # 可平空数量
    available_short = Column(Float)
    # 平均做空价格
    average_short_price = Column(Float)

    # 盈亏
    profit = Column(Float)
    # 盈亏比例
    profit_rate = Column(Float)
    # 市值 或者 占用的保证金(方便起见，总是100%)
    value = Column(Float)
    # 交易类型(0代表T+0,1代表T+1)
    trading_t = Column(Integer)


# 委托单
class Order(TraderBase, Mixin):
    __tablename__ = 'orders'

    # 机器人名字
    trader_name = Column(String(length=128))
    # 订单价格
    order_price = Column(Float)
    # 订单数量
    order_amount = Column(Float)
    # 订单类型
    order_type = Column(String(length=64))
    # 订单状态
    status = Column(String(length=64))

    # 产生订单的selector/factor level
    level = Column(String(length=32))


register_schema(providers=['zvt'], db_name='trader_info', schema_base=TraderBase)

# the __all__ is generated
__all__ = ['TraderInfo', 'AccountStats', 'Position', 'Order']


# -*- coding: utf-8 -*-
from typing import List

import pandas as pd

from zvtm.api import get_kdata_schema
from zvtm.contract.api import decode_entity_id
from zvtm.contract.drawer import Drawer, ChartType
from zvtm.domain import Index1dKdata


def compare(entity_ids, columns=None, chart_type: ChartType = ChartType.line):
    entity_type_map_ids = _group_entity_ids(entity_ids=entity_ids)
    # compare kdata
    if columns is None:
        dfs = []
        for entity_type in entity_type_map_ids:
            schema = get_kdata_schema(entity_type=entity_type)
            df = schema.query_data(entity_ids=entity_type_map_ids.get(entity_type))
            dfs.append(df)
        all_df = pd.concat(dfs)
        drawer = Drawer(main_df=all_df, sub_df_list=[all_df[['entity_id', 'timestamp', 'turnover']].copy()])
        drawer.draw_kline(main_chart=chart_type, show=True)


def distribute(entity_ids, data_schema, columns, histnorm='percent', nbinsx=20, filters=None):
    df = data_schema.query_data(entity_ids=entity_ids, columns=columns, filters=filters)
    drawer = Drawer(main_df=df)
    drawer.draw_histogram(show=True, histnorm=histnorm, nbinsx=nbinsx)


def _group_entity_ids(entity_ids):
    entity_type_map_ids = {}
    for entity_id in entity_ids:
        entity_type, _, _ = decode_entity_id(entity_id)
        ids: List = entity_type_map_ids.setdefault(entity_type, [])
        ids.append(entity_id)
    return entity_type_map_ids


if __name__ == '__main__':
    #                  id        entity_id  timestamp entity_type exchange    code  name  list_date end_date publisher category  base_point
    # 16  index_sz_399370  index_sz_399370 2002-12-31       index       sz  399370  国证成长 2010-01-04     None   cnindex    style      1000.0
    # 17  index_sz_399371  index_sz_399371 2002-12-31       index       sz  399371  国证价值 2010-01-04     None   cnindex    style      1000.0
    # 18  index_sz_399372  index_sz_399372 2002-12-31       index       sz  399372  大盘成长 2010-01-04     None   cnindex    style      1000.0
    # 19  index_sz_399373  index_sz_399373 2002-12-31       index       sz  399373  大盘价值 2010-01-04     None   cnindex    style      1000.0
    # 20  index_sz_399374  index_sz_399374 2002-12-31       index       sz  399374  中盘成长 2010-01-04     None   cnindex    style      1000.0
    # 21  index_sz_399375  index_sz_399375 2002-12-31       index       sz  399375  中盘价值 2010-01-04     None   cnindex    style      1000.0
    # 22  index_sz_399376  index_sz_399376 2002-12-31       index       sz  399376  小盘成长 2010-01-04     None   cnindex    style      1000.0
    # 23  index_sz_399377  index_sz_399377 2002-12-31       index       sz  399377  小盘价值 2010-01-04     None   cnindex    style      1000.0

    # 成长 大 中 小
    # entity_ids = ['index_sz_399372', 'index_sz_399374', 'index_sz_399376']

    # 价值 大 中 小
    entity_ids = ['index_sz_399373', 'index_sz_399375', 'index_sz_399377']

    # Index1dKdata.record_data(entity_ids=entity_ids)
    # compare(entity_ids=entity_ids)

    # df1 = Index1dKdata.query_data(entity_id='index_sz_399370', index=['timestamp'],
    #                               columns=['entity_id', 'timestamp', 'close'])
    # df2 = Index1dKdata.query_data(entity_id='index_sz_399371', index=['timestamp'],
    #                               columns=['entity_id', 'timestamp', 'close'])
    # df = (df1['close'] / df2['close']).to_frame()
    # df['entity_id'] = '399370 / 399371'
    # df = df.reset_index()

    distribute(entity_ids=['index_sh_000001', 'index_sz_399001'], columns=['entity_id', 'timestamp', 'turnover_rate'],
               data_schema=Index1dKdata, filters=[Index1dKdata.turnover_rate > 0], nbinsx=10)



# -*- coding: utf-8 -*-
import enum
from typing import Union, Type

import pandas as pd

from zvtm.api import get_kdata_schema
from zvtm.contract import IntervalLevel, AdjustType
from zvtm.contract import TradableEntity
from zvtm.domain import Stock
from zvtm.tag.dataset.stock_tags import StockTags
from zvtm.utils import next_date, to_pd_timestamp


class RelativePerformance(enum.Enum):
    # 表现比90%好
    best = 0.9
    ordinary = 0.5
    poor = 0


def cal_change(s):
    return (s[-1] - s[0]) / s[0]


def cal_performance(s):
    if s >= RelativePerformance.best.value:
        return RelativePerformance.best.value
    if s >= RelativePerformance.ordinary.value:
        return RelativePerformance.ordinary.value
    if s >= RelativePerformance.poor.value:
        return RelativePerformance.poor.value


def get_performance(x_timestamp,
                    entity_type='stock',
                    entity_ids=None,
                    predict_range=20,
                    level: Union[IntervalLevel, str] = IntervalLevel.LEVEL_1DAY,
                    adjust_type: Union[AdjustType, str] = None):
    kdata_schema = get_kdata_schema(entity_type=entity_type, level=level, adjust_type=adjust_type)
    y_df = kdata_schema.query_data(start_timestamp=x_timestamp,
                                   end_timestamp=next_date(x_timestamp, predict_range), entity_ids=entity_ids,
                                   columns=['entity_id', 'timestamp', 'close'],
                                   index=['entity_id', 'timestamp'])
    y_df = y_df.dropna()
    y_change = y_df.groupby(level=0)['close'].apply(
        lambda x: cal_change(x)).rename('y_change')
    y_score = y_change.rank(pct=True).apply(
        lambda x: cal_performance(x)).rename('y_score')

    df = y_score.to_frame()
    df['timestamp'] = x_timestamp
    df.set_index('timestamp', append=True)

    return df


def get_samples(data_schema,
                columns,
                x_timestamp,
                entity_type='stock',
                entity_ids=None,
                predict_range=20,
                level: Union[IntervalLevel, str] = IntervalLevel.LEVEL_1DAY,
                adjust_type: Union[AdjustType, str] = None):
    # features
    x_df = data_schema.query_data(start_timestamp=x_timestamp, end_timestamp=x_timestamp, entity_ids=entity_ids,
                                  columns=columns, index=['entity_id'])
    x_df = x_df.dropna()

    # Y
    y = get_performance(x_timestamp=x_timestamp, entity_type=entity_type, entity_ids=entity_ids,
                        predict_range=predict_range, level=level, adjust_type=adjust_type)
    index = x_df.index & y.index
    print(index)
    return x_df.loc[index].to_numpy(), y.loc[index].to_numpy().tolist()


class MLMachine(object):
    entity_schema: Type[TradableEntity] = None

    sample_start_timestamp = '2005-01-01'

    test_start_timestamp = '2010-01-01'

    def __init__(self, entity_ids=None, predict_range=20, level: Union[IntervalLevel, str] = IntervalLevel.LEVEL_1DAY,
                 adjust_type: Union[AdjustType, str] = None) -> None:
        super().__init__()
        self.entity_ids = entity_ids
        self.predict_range = predict_range
        self.level = level
        if not adjust_type and self.entity_schema == Stock:
            self.adjust_type = AdjustType.hfq
        else:
            self.adjust_type = adjust_type

        self.sample_start_timestamp = to_pd_timestamp(self.sample_start_timestamp)
        self.test_start_timestamp = to_pd_timestamp(self.test_start_timestamp)

        self.x_timestamps, self.y_timestamps = self.get_x_y_timestamps()

        self.x_df = self.get_features(self.entity_ids, self.x_timestamps)
        self.y_df = self.get_labels(self.entity_ids, x_timestamps=self.x_timestamps, y_timestamps=self.y_timestamps)

    def ml(self):
        print(self.x_df)
        print(self.y_df)

    def get_x_y_timestamps(self):
        x_timestamps = []
        y_timestamps = []
        x_timestamp = self.sample_start_timestamp
        y_timestamp = next_date(x_timestamp, self.predict_range)
        while y_timestamp <= self.test_start_timestamp:
            x_timestamps.append(x_timestamp)
            y_timestamps.append(y_timestamp)

            x_timestamp = y_timestamp
            y_timestamp = next_date(x_timestamp, self.predict_range)

        return y_timestamps, y_timestamps

    def get_features(self, entity_ids, timestamps):
        raise NotImplementedError

    def get_labels(self, entity_ids, x_timestamps, y_timestamps):
        dfs = []
        for idx, timestamp in enumerate(x_timestamps):
            kdata_schema = get_kdata_schema(entity_type=self.entity_schema.__name__.lower(), level=self.level,
                                            adjust_type=self.adjust_type)
            y_df = kdata_schema.query_data(start_timestamp=timestamp,
                                           end_timestamp=y_timestamps[idx],
                                           entity_ids=entity_ids,
                                           columns=['entity_id', 'timestamp', 'close'],
                                           index=['entity_id', 'timestamp'])
            y_df = y_df.dropna()
            y_change = y_df.groupby(level=0)['close'].apply(
                lambda x: cal_change(x)).rename('y_change')
            y_score = y_change.rank(pct=True).apply(
                lambda x: cal_performance(x)).rename('y_score')

            df = y_score.to_frame()
            df['timestamp'] = timestamp
            df.set_index('timestamp', append=True)
            dfs.append(df)

        return pd.concat(dfs)

    def make_samples(self):
        pass


class MyMLMachine(MLMachine):
    entity_schema = Stock

    def get_features(self, entity_ids, timestamps):
        return StockTags.query_data(columns=['entity_id', 'timestamp', 'cycle_tag'],
                                    filters=[StockTags.timestamp.in_(timestamps)])


if __name__ == '__main__':
    MyMLMachine().ml()



# -*- coding: utf-8 -*-
from typing import List

import pandas as pd

from zvtm.api.utils import to_report_period_type, value_to_pct
from zvtm.contract import ActorType
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import TimestampsDataRecorder
from zvtm.domain import Stock, ActorMeta
from zvtm.domain.actor.stock_actor import StockTopTenFreeHolder, StockInstitutionalInvestorHolder
from zvtm.recorders.em.em_api import get_holder_report_dates, get_free_holders
from zvtm.utils import to_pd_timestamp, to_time_str


class EMStockTopTenFreeRecorder(TimestampsDataRecorder):
    entity_provider = 'joinquant'
    entity_schema = Stock

    provider = 'em'
    data_schema = StockTopTenFreeHolder

    def init_timestamps(self, entity_item) -> List[pd.Timestamp]:
        result = get_holder_report_dates(code=entity_item.code)
        if result:
            return [to_pd_timestamp(item['END_DATE']) for item in result]

    def on_finish_entity(self, entity):
        super().on_finish_entity(entity)
        holders = StockTopTenFreeHolder.query_data(entity_id=entity.id,
                                                   filters=[StockTopTenFreeHolder.holding_values == None],
                                                   session=self.session,
                                                   return_type='domain')
        for holder in holders:
            ii = StockInstitutionalInvestorHolder.query_data(entity_id=entity.id,
                                                             filters=[
                                                                 StockInstitutionalInvestorHolder.holding_values > 1,
                                                                 StockInstitutionalInvestorHolder.holding_ratio > 0.01,
                                                                 StockInstitutionalInvestorHolder.timestamp == holder.timestamp],
                                                             limit=1,
                                                             return_type='domain')
            if ii:
                holder.holding_values = holder.holding_ratio * ii[0].holding_values / ii[0].holding_ratio
        self.session.commit()

    def record(self, entity, start, end, size, timestamps):
        for timestamp in timestamps:
            the_date = to_time_str(timestamp)
            result = get_free_holders(code=entity.code, end_date=the_date)
            if result:
                holders = []
                new_actors = []
                for item in result:
                    # {'END_DATE': '2021-03-31 00:00:00',
                    #   'FREE_HOLDNUM_RATIO': 0.631949916991,
                    #   'FREE_RATIO_QOQ': '-5.33046217',
                    #   'HOLDER_CODE': '161606',
                    #   'HOLDER_CODE_OLD': '161606',
                    #   'HOLDER_NAME': '交通银行-融通行业景气证券投资基金',
                    #   'HOLDER_RANK': 10,
                    #   'HOLD_NUM': 39100990,
                    #   'IS_HOLDORG': '1',
                    #   'SECUCODE': '000338.SZ'}
                    # 机构
                    if item['IS_HOLDORG'] == '1':
                        domains: List[ActorMeta] = ActorMeta.query_data(filters=[ActorMeta.code == item['HOLDER_CODE']],
                                                                        return_type='domain')
                        if not domains:
                            actor_type = ActorType.corporation.value
                            actor = ActorMeta(entity_id=f'{actor_type}_cn_{item["HOLDER_CODE"]}',
                                              id=f'{actor_type}_cn_{item["HOLDER_CODE"]}',
                                              entity_type=actor_type,
                                              exchange='cn',
                                              code=item["HOLDER_CODE"],
                                              name=item["HOLDER_NAME"])
                        else:
                            actor = domains[0]
                    else:
                        actor_type = ActorType.individual.value
                        actor = ActorMeta(entity_id=f'{actor_type}_cn_{item["HOLDER_NAME"]}',
                                          id=f'{actor_type}_cn_{item["HOLDER_NAME"]}',
                                          entity_type=actor_type,
                                          exchange='cn',
                                          code=item["HOLDER_NAME"],
                                          name=item["HOLDER_NAME"])
                        new_actors.append(actor.__dict__)
                    holder = {'id': f'{entity.entity_id}_{the_date}_{actor.entity_id}',
                              'entity_id': entity.entity_id,
                              'timestamp': timestamp,
                              'code': entity.code,
                              'name': entity.name,

                              'actor_id': actor.entity_id,
                              'actor_type': actor.entity_type,
                              'actor_code': actor.code,
                              'actor_name': actor.name,

                              'report_date': timestamp,
                              'report_period': to_report_period_type(timestamp),

                              'holding_numbers': item['HOLD_NUM'],
                              'holding_ratio': value_to_pct(item['FREE_HOLDNUM_RATIO'], 0)}
                    holders.append(holder)
                if holders:
                    df = pd.DataFrame.from_records(holders)
                    df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                             force_update=True)
                if new_actors:
                    df = pd.DataFrame.from_records(new_actors)
                    df_to_db(data_schema=ActorMeta, df=df, provider=self.provider,
                             force_update=False)


if __name__ == '__main__':
    EMStockTopTenFreeRecorder(codes=['000338']).run()
# the __all__ is generated
__all__ = ['EMStockTopTenFreeRecorder']


# -*- coding: utf-8 -*-

import io
import re

import demjson
import pandas as pd
import requests

from zvtm.api.utils import china_stock_code_to_id
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import Recorder
from zvtm.domain import EtfStock, Etf
from zvtm.recorders.consts import DEFAULT_SH_ETF_LIST_HEADER
from zvtm.utils.time_utils import now_pd_timestamp


class ChinaETFListSpider(Recorder):
    data_schema = EtfStock

    def __init__(self, force_update=False, sleeping_time=10.0, provider='exchange') -> None:
        self.provider = provider
        super().__init__(force_update, sleeping_time)

    def run(self):
        # 抓取沪市 ETF 列表
        url = 'http://query.sse.com.cn/commonQuery.do?sqlId=COMMON_SSE_ZQPZ_ETFLB_L_NEW'
        response = requests.get(url, headers=DEFAULT_SH_ETF_LIST_HEADER)
        response_dict = demjson.decode(response.text)

        df = pd.DataFrame(response_dict.get('result', []))
        self.persist_etf_list(df, exchange='sh')
        self.logger.info('沪市 ETF 列表抓取完成...')

        # 抓取沪市 ETF 成分股
        self.download_sh_etf_component(df)
        self.logger.info('沪市 ETF 成分股抓取完成...')

        # 抓取深市 ETF 列表
        url = 'http://www.szse.cn/api/report/ShowReport?SHOWTYPE=xlsx&CATALOGID=1945'
        response = requests.get(url)

        df = pd.read_excel(io.BytesIO(response.content), dtype=str)
        self.persist_etf_list(df, exchange='sz')
        self.logger.info('深市 ETF 列表抓取完成...')

        # 抓取深市 ETF 成分股
        self.download_sz_etf_component(df)
        self.logger.info('深市 ETF 成分股抓取完成...')

    def persist_etf_list(self, df: pd.DataFrame, exchange: str):
        if df is None:
            return

        df = df.copy()
        if exchange == 'sh':
            df = df[['FUND_ID', 'FUND_NAME']]
        elif exchange == 'sz':
            df = df[['证券代码', '证券简称']]

        df.columns = ['code', 'name']
        df['id'] = df['code'].apply(lambda code: f'etf_{exchange}_{code}')
        df['entity_id'] = df['id']
        df['exchange'] = exchange
        df['entity_type'] = 'etf'
        df['category'] = 'etf'

        df = df.dropna(axis=0, how='any')
        df = df.drop_duplicates(subset='id', keep='last')

        df_to_db(df=df, data_schema=Etf, provider=self.provider, force_update=False)

    def download_sh_etf_component(self, df: pd.DataFrame):
        """
        ETF_CLASS => 1. 单市场 ETF 2.跨市场 ETF 3. 跨境 ETF
                        5. 债券 ETF 6. 黄金 ETF
        :param df: ETF 列表数据
        :return: None
        """
        query_url = 'http://query.sse.com.cn/infodisplay/queryConstituentStockInfo.do?' \
                    'isPagination=false&type={}&etfClass={}'

        etf_df = df[(df['ETF_CLASS'] == '1') | (df['ETF_CLASS'] == '2')]
        etf_df = self.populate_sh_etf_type(etf_df)

        for _, etf in etf_df.iterrows():
            url = query_url.format(etf['ETF_TYPE'], etf['ETF_CLASS'])
            response = requests.get(url, headers=DEFAULT_SH_ETF_LIST_HEADER)
            response_dict = demjson.decode(response.text)
            response_df = pd.DataFrame(response_dict.get('result', []))

            etf_code = etf['FUND_ID']
            etf_id = f'etf_sh_{etf_code}'
            response_df = response_df[['instrumentId', 'instrumentName']].copy()
            response_df.rename(columns={'instrumentId': 'stock_code', 'instrumentName': 'stock_name'}, inplace=True)

            response_df['entity_id'] = etf_id
            response_df['entity_type'] = 'etf'
            response_df['exchange'] = 'sh'
            response_df['code'] = etf_code
            response_df['name'] = etf['FUND_NAME']
            response_df['timestamp'] = now_pd_timestamp()

            response_df['stock_id'] = response_df['stock_code'].apply(lambda code: china_stock_code_to_id(code))
            response_df['id'] = response_df['stock_id'].apply(
                lambda x: f'{etf_id}_{x}')

            df_to_db(data_schema=self.data_schema, df=response_df, provider=self.provider)
            self.logger.info(f'{etf["FUND_NAME"]} - {etf_code} 成分股抓取完成...')

            self.sleep()

    def download_sz_etf_component(self, df: pd.DataFrame):
        query_url = 'http://vip.stock.finance.sina.com.cn/corp/go.php/vII_NewestComponent/indexid/{}.phtml'

        self.parse_sz_etf_underlying_index(df)
        for _, etf in df.iterrows():
            underlying_index = etf['拟合指数']
            etf_code = etf['证券代码']

            if len(underlying_index) == 0:
                self.logger.info(f'{etf["证券简称"]} - {etf_code} 非 A 股市场指数，跳过...')
                continue

            url = query_url.format(underlying_index)
            response = requests.get(url)
            response.encoding = 'gbk'

            try:
                dfs = pd.read_html(response.text, header=1)
            except ValueError as error:
                self.logger.error(f'HTML parse error: {error}, response: {response.text}')
                continue

            if len(dfs) < 4:
                continue

            response_df = dfs[3].copy()
            response_df = response_df.dropna(axis=1, how='any')
            response_df['品种代码'] = response_df['品种代码'].apply(lambda x: f'{x:06d}')

            etf_id = f'etf_sz_{etf_code}'
            response_df = response_df[['品种代码', '品种名称']].copy()
            response_df.rename(columns={'品种代码': 'stock_code', '品种名称': 'stock_name'}, inplace=True)

            response_df['entity_id'] = etf_id
            response_df['entity_type'] = 'etf'
            response_df['exchange'] = 'sz'
            response_df['code'] = etf_code
            response_df['name'] = etf['证券简称']
            response_df['timestamp'] = now_pd_timestamp()

            response_df['stock_id'] = response_df['stock_code'].apply(lambda code: china_stock_code_to_id(code))
            response_df['id'] = response_df['stock_id'].apply(
                lambda x: f'{etf_id}_{x}')

            df_to_db(data_schema=self.data_schema, df=response_df, provider=self.provider)
            self.logger.info(f'{etf["证券简称"]} - {etf_code} 成分股抓取完成...')

            self.sleep()

    @staticmethod
    def populate_sh_etf_type(df: pd.DataFrame):
        """
        填充沪市 ETF 代码对应的 TYPE 到列表数据中
        :param df: ETF 列表数据
        :return: 包含 ETF 对应 TYPE 的列表数据
        """
        query_url = 'http://query.sse.com.cn/infodisplay/queryETFNewAllInfo.do?' \
                    'isPagination=false&type={}&pageHelp.pageSize=25'

        type_df = pd.DataFrame()
        for etf_class in [1, 2]:
            url = query_url.format(etf_class)
            response = requests.get(url, headers=DEFAULT_SH_ETF_LIST_HEADER)
            response_dict = demjson.decode(response.text)
            response_df = pd.DataFrame(response_dict.get('result', []))
            response_df = response_df[['fundid1', 'etftype']]

            type_df = pd.concat([type_df, response_df])

        result_df = df.copy()
        result_df = result_df.sort_values(by='FUND_ID').reset_index(drop=True)
        type_df = type_df.sort_values(by='fundid1').reset_index(drop=True)

        result_df['ETF_TYPE'] = type_df['etftype']

        return result_df

    @staticmethod
    def parse_sz_etf_underlying_index(df: pd.DataFrame):
        """
        解析深市 ETF 对应跟踪的指数代码
        :param df: ETF 列表数据
        :return: 解析完成 ETF 对应指数代码的列表数据
        """

        def parse_index(text):
            if len(text) == 0:
                return ''

            result = re.search(r"(\d+).*", text)
            if result is None:
                return ''
            else:
                return result.group(1)

        df['拟合指数'] = df['拟合指数'].apply(parse_index)


__all__ = ['ChinaETFListSpider']

if __name__ == '__main__':
    spider = ChinaETFListSpider(provider='exchange')
    spider.run()
# the __all__ is generated
__all__ = ['ChinaETFListSpider']


# -*- coding: utf-8 -*-

from zvtm.utils.pd_utils import pd_is_not_null, fill_with_same_index, normal_index_df, is_normal_df


class NormalData(object):
    table_type_sample = None

    def __init__(self,
                 df,
                 category_field='entity_id',
                 time_field='timestamp',
                 fill_index: bool = False) -> None:
        self.data_df = df
        self.category_field = category_field
        self.time_field = time_field
        self.fill_index = fill_index

        self.entity_ids = []
        self.df_list = []
        self.entity_map_df = {}

        self.normalize()

    def normalize(self):
        """
        normalize data_df to
                                    col1    col2    col3
        entity_id    timestamp

        """
        if pd_is_not_null(self.data_df):
            if not is_normal_df(self.data_df):
                self.data_df = normal_index_df(self.data_df, self.category_field, self.time_field)

            self.entity_ids = self.data_df.index.levels[0].to_list()

            for entity_id in self.entity_ids:
                df = self.data_df.loc[(entity_id,)]
                self.df_list.append(df)
                self.entity_map_df[entity_id] = df

            if len(self.df_list) > 1 and self.fill_index:
                self.df_list = fill_with_same_index(df_list=self.df_list)

    def empty(self):
        return not pd_is_not_null(self.data_df)

# the __all__ is generated
__all__ = ['NormalData']


# -*- coding: utf-8 -*-
from sqlalchemy import Column, String, Text
from sqlalchemy.orm import declarative_base

from zvtm.contract import Mixin
from zvtm.contract.register import register_schema

ZvtInfoBase = declarative_base()


# 用于保存recorder的状态
class RecorderState(ZvtInfoBase, Mixin):
    __tablename__ = 'recoder_state'
    # recorder名字
    recoder_name = Column(String(length=128))

    # json string
    state = Column(Text())


# 用于保存recorder的状态
class TaggerState(ZvtInfoBase, Mixin):
    __tablename__ = 'tagger_state'
    # tagger名字
    tagger_name = Column(String(length=128))

    # json string
    state = Column(Text())


register_schema(providers=['zvt'], db_name='zvt_info', schema_base=ZvtInfoBase)
# the __all__ is generated
__all__ = ['RecorderState', 'TaggerState']



# -*- coding: utf-8 -*-
from sqlalchemy import Column, String, DateTime, Float
from sqlalchemy.orm import declarative_base

from zvtm.contract import Mixin
from zvtm.contract.register import register_schema

HolderBase = declarative_base()


class HkHolder(HolderBase, Mixin):
    __tablename__ = 'hk_holder'
    # 股票代码
    code = Column(String(length=32))
    # 股票名称
    name = Column(String(length=32))

    # 市场通编码	三种类型：310001-沪股通，310002-深股通，310005-港股通
    holder_code = Column(String(length=32))
    # 市场通名称	三种类型：沪股通，深股通，港股通
    holder_name = Column(String(length=128))

    # 持股数量
    share_number = Column(Float)
    # 持股比例
    share_ratio = Column(Float)


class TopTenTradableHolder(HolderBase, Mixin):
    __tablename__ = 'top_ten_tradable_holder'

    provider = Column(String(length=32))
    code = Column(String(length=32))

    report_period = Column(String(length=32))
    report_date = Column(DateTime)

    # 股东代码
    holder_code = Column(String(length=32))
    # 股东名称
    holder_name = Column(String(length=128))
    # 持股数
    shareholding_numbers = Column(Float)
    # 持股比例
    shareholding_ratio = Column(Float)
    # 变动
    change = Column(Float)
    # 变动比例
    change_ratio = Column(Float)


class TopTenHolder(HolderBase, Mixin):
    __tablename__ = 'top_ten_holder'

    provider = Column(String(length=32))
    code = Column(String(length=32))

    report_period = Column(String(length=32))
    report_date = Column(DateTime)

    # 股东代码
    holder_code = Column(String(length=32))
    # 股东名称
    holder_name = Column(String(length=128))
    # 持股数
    shareholding_numbers = Column(Float)
    # 持股比例
    shareholding_ratio = Column(Float)
    # 变动
    change = Column(Float)
    # 变动比例
    change_ratio = Column(Float)


class InstitutionalInvestorHolder(HolderBase, Mixin):
    __tablename__ = 'institutional_investor_holder'

    provider = Column(String(length=32))
    code = Column(String(length=32))

    report_period = Column(String(length=32))
    report_date = Column(DateTime)

    # 机构类型
    institutional_investor_type = Column(String(length=64))
    # 股东代码
    holder_code = Column(String(length=32))
    # 股东名称
    holder_name = Column(String(length=128))
    # 持股数
    shareholding_numbers = Column(Float)
    # 持股比例
    shareholding_ratio = Column(Float)


register_schema(providers=['eastmoney', 'joinquant'], db_name='holder', schema_base=HolderBase, entity_type='stock')

# the __all__ is generated
__all__ = ['HkHolder', 'TopTenTradableHolder', 'TopTenHolder', 'InstitutionalInvestorHolder']


# -*- coding: utf-8 -*-

from jqdatapy.api import get_fundamentals

from zvtm.api.utils import to_report_period_type
from zvtm.contract.api import get_data
from zvtm.domain import FinanceFactor, ReportPeriod
from zvtm.recorders.eastmoney.common import company_type_flag, get_fc, EastmoneyTimestampsDataRecorder, \
    call_eastmoney_api, get_from_path_fields
from zvtm.recorders.joinquant.common import to_jq_entity_id
from zvtm.utils.pd_utils import index_df
from zvtm.utils.pd_utils import pd_is_not_null
from zvtm.utils.time_utils import to_time_str, to_pd_timestamp


def to_jq_report_period(timestamp):
    the_date = to_pd_timestamp(timestamp)
    report_period = to_report_period_type(timestamp)
    if report_period == ReportPeriod.year.value:
        return '{}'.format(the_date.year)
    if report_period == ReportPeriod.season1.value:
        return '{}q1'.format(the_date.year)
    if report_period == ReportPeriod.half_year.value:
        return '{}q2'.format(the_date.year)
    if report_period == ReportPeriod.season3.value:
        return '{}q3'.format(the_date.year)

    assert False


class BaseChinaStockFinanceRecorder(EastmoneyTimestampsDataRecorder):
    finance_report_type = None
    data_type = 1

    timestamps_fetching_url = 'https://emh5.eastmoney.com/api/CaiWuFenXi/GetCompanyReportDateList'
    timestamp_list_path_fields = ['CompanyReportDateList']
    timestamp_path_fields = ['ReportDate']

    def __init__(self, exchanges=None, entity_ids=None, code=None, codes=None, day_data=False,
                 force_update=False, sleeping_time=5, real_time=False,
                 fix_duplicate_way='add', start_timestamp=None, end_timestamp=None) -> None:
        super().__init__(force_update, sleeping_time, exchanges, entity_ids, code, codes, day_data, real_time=real_time,
                         fix_duplicate_way=fix_duplicate_way, start_timestamp=start_timestamp,
                         end_timestamp=end_timestamp)

        try:
            self.fetch_jq_timestamp = True
        except Exception as e:
            self.fetch_jq_timestamp = False
            self.logger.warning(
                f'joinquant account not ok,the timestamp(publish date) for finance would be not correct', e)

    def init_timestamps(self, entity):
        param = {
            "color": "w",
            "fc": get_fc(entity),
            "DataType": self.data_type
        }

        if self.finance_report_type == 'LiRunBiaoList' or self.finance_report_type == 'XianJinLiuLiangBiaoList':
            param['ReportType'] = 1

        timestamp_json_list = call_eastmoney_api(url=self.timestamps_fetching_url,
                                                 path_fields=self.timestamp_list_path_fields,
                                                 param=param)

        if self.timestamp_path_fields:
            timestamps = [get_from_path_fields(data, self.timestamp_path_fields) for data in timestamp_json_list]

        return [to_pd_timestamp(t) for t in timestamps]

    def generate_request_param(self, security_item, start, end, size, timestamps):
        if len(timestamps) <= 10:
            param = {
                "color": "w",
                "fc": get_fc(security_item),
                "corpType": company_type_flag(security_item),
                # 0 means get all types
                "reportDateType": 0,
                "endDate": '',
                "latestCount": size
            }
        else:
            param = {
                "color": "w",
                "fc": get_fc(security_item),
                "corpType": company_type_flag(security_item),
                # 0 means get all types
                "reportDateType": 0,
                "endDate": to_time_str(timestamps[10]),
                "latestCount": 10
            }

        if self.finance_report_type == 'LiRunBiaoList' or self.finance_report_type == 'XianJinLiuLiangBiaoList':
            param['reportType'] = 1

        return param

    def generate_path_fields(self, security_item):
        comp_type = company_type_flag(security_item)

        if comp_type == "3":
            return ['{}_YinHang'.format(self.finance_report_type)]
        elif comp_type == "2":
            return ['{}_BaoXian'.format(self.finance_report_type)]
        elif comp_type == "1":
            return ['{}_QuanShang'.format(self.finance_report_type)]
        elif comp_type == "4":
            return ['{}_QiYe'.format(self.finance_report_type)]

    def record(self, entity, start, end, size, timestamps):
        # different with the default timestamps handling
        param = self.generate_request_param(entity, start, end, size, timestamps)
        self.logger.info('request param:{}'.format(param))

        return self.api_wrapper.request(url=self.url, param=param, method=self.request_method,
                                        path_fields=self.generate_path_fields(entity))

    def get_original_time_field(self):
        return 'ReportDate'

    def fill_timestamp_with_jq(self, security_item, the_data):
        # get report published date from jq
        try:
            df = get_fundamentals(table='indicator', code=to_jq_entity_id(security_item), columns='pubDate',
                                  date=to_jq_report_period(the_data.report_date), count=None, parse_dates=['pubDate'])
            if pd_is_not_null(df):
                the_data.timestamp = to_pd_timestamp(df['pubDate'][0])
                self.logger.info(
                    'jq fill {} {} timestamp:{} for report_date:{}'.format(self.data_schema, security_item.id,
                                                                           the_data.timestamp,
                                                                           the_data.report_date))
                self.session.commit()
        except Exception as e:
            self.logger.error(e)

    def on_finish_entity(self, entity):
        super().on_finish_entity(entity)

        if not self.fetch_jq_timestamp:
            return

        # fill the timestamp for report published date
        the_data_list = get_data(data_schema=self.data_schema,
                                 provider=self.provider,
                                 entity_id=entity.id,
                                 order=self.data_schema.timestamp.asc(),
                                 return_type='domain',
                                 session=self.session,
                                 filters=[self.data_schema.timestamp == self.data_schema.report_date,
                                          self.data_schema.timestamp >= to_pd_timestamp('2005-01-01')])
        if the_data_list:
            if self.data_schema == FinanceFactor:
                for the_data in the_data_list:
                    self.fill_timestamp_with_jq(entity, the_data)
            else:
                df = FinanceFactor.query_data(entity_id=entity.id,
                                              columns=[FinanceFactor.timestamp, FinanceFactor.report_date,
                                                       FinanceFactor.id],
                                              filters=[FinanceFactor.timestamp != FinanceFactor.report_date,
                                                       FinanceFactor.timestamp >= to_pd_timestamp('2005-01-01'),
                                                       FinanceFactor.report_date >= the_data_list[0].report_date,
                                                       FinanceFactor.report_date <= the_data_list[-1].report_date, ])

                if pd_is_not_null(df):
                    index_df(df, index='report_date', time_field='report_date')

                for the_data in the_data_list:
                    if (df is not None) and (not df.empty) and the_data.report_date in df.index:
                        the_data.timestamp = df.at[the_data.report_date, 'timestamp']
                        self.logger.info(
                            'db fill {} {} timestamp:{} for report_date:{}'.format(self.data_schema, entity.id,
                                                                                   the_data.timestamp,
                                                                                   the_data.report_date))
                        self.session.commit()
                    else:
                        # self.logger.info(
                        #     'waiting jq fill {} {} timestamp:{} for report_date:{}'.format(self.data_schema,
                        #                                                                    security_item.id,
                        #                                                                    the_data.timestamp,
                        #                                                                    the_data.report_date))

                        self.fill_timestamp_with_jq(entity, the_data)


# the __all__ is generated
__all__ = ['to_jq_report_period', 'BaseChinaStockFinanceRecorder']


# -*- coding: utf-8 -*-
# this file is generated by gen_kdata_schema function, dont't change it
# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule stock_1h_kdata
from .stock_1h_kdata import *
from .stock_1h_kdata import __all__ as _stock_1h_kdata_all
__all__ += _stock_1h_kdata_all

# import all from submodule stock_15m_hfq_kdata
from .stock_15m_hfq_kdata import *
from .stock_15m_hfq_kdata import __all__ as _stock_15m_hfq_kdata_all
__all__ += _stock_15m_hfq_kdata_all

# import all from submodule stock_1wk_kdata
from .stock_1wk_kdata import *
from .stock_1wk_kdata import __all__ as _stock_1wk_kdata_all
__all__ += _stock_1wk_kdata_all

# import all from submodule stock_15m_kdata
from .stock_15m_kdata import *
from .stock_15m_kdata import __all__ as _stock_15m_kdata_all
__all__ += _stock_15m_kdata_all

# import all from submodule stock_1m_hfq_kdata
from .stock_1m_hfq_kdata import *
from .stock_1m_hfq_kdata import __all__ as _stock_1m_hfq_kdata_all
__all__ += _stock_1m_hfq_kdata_all

# import all from submodule stock_4h_hfq_kdata
from .stock_4h_hfq_kdata import *
from .stock_4h_hfq_kdata import __all__ as _stock_4h_hfq_kdata_all
__all__ += _stock_4h_hfq_kdata_all

# import all from submodule stock_5m_hfq_kdata
from .stock_5m_hfq_kdata import *
from .stock_5m_hfq_kdata import __all__ as _stock_5m_hfq_kdata_all
__all__ += _stock_5m_hfq_kdata_all

# import all from submodule stock_5m_kdata
from .stock_5m_kdata import *
from .stock_5m_kdata import __all__ as _stock_5m_kdata_all
__all__ += _stock_5m_kdata_all

# import all from submodule stock_1d_kdata
from .stock_1d_kdata import *
from .stock_1d_kdata import __all__ as _stock_1d_kdata_all
__all__ += _stock_1d_kdata_all

# import all from submodule stock_30m_hfq_kdata
from .stock_30m_hfq_kdata import *
from .stock_30m_hfq_kdata import __all__ as _stock_30m_hfq_kdata_all
__all__ += _stock_30m_hfq_kdata_all

# import all from submodule stock_1mon_hfq_kdata
from .stock_1mon_hfq_kdata import *
from .stock_1mon_hfq_kdata import __all__ as _stock_1mon_hfq_kdata_all
__all__ += _stock_1mon_hfq_kdata_all

# import all from submodule stock_1wk_hfq_kdata
from .stock_1wk_hfq_kdata import *
from .stock_1wk_hfq_kdata import __all__ as _stock_1wk_hfq_kdata_all
__all__ += _stock_1wk_hfq_kdata_all

# import all from submodule stock_1mon_kdata
from .stock_1mon_kdata import *
from .stock_1mon_kdata import __all__ as _stock_1mon_kdata_all
__all__ += _stock_1mon_kdata_all

# import all from submodule stock_1h_hfq_kdata
from .stock_1h_hfq_kdata import *
from .stock_1h_hfq_kdata import __all__ as _stock_1h_hfq_kdata_all
__all__ += _stock_1h_hfq_kdata_all

# import all from submodule stock_1m_kdata
from .stock_1m_kdata import *
from .stock_1m_kdata import __all__ as _stock_1m_kdata_all
__all__ += _stock_1m_kdata_all

# import all from submodule stock_4h_kdata
from .stock_4h_kdata import *
from .stock_4h_kdata import __all__ as _stock_4h_kdata_all
__all__ += _stock_4h_kdata_all

# import all from submodule stock_1d_hfq_kdata
from .stock_1d_hfq_kdata import *
from .stock_1d_hfq_kdata import __all__ as _stock_1d_hfq_kdata_all
__all__ += _stock_1d_hfq_kdata_all

# import all from submodule stock_30m_kdata
from .stock_30m_kdata import *
from .stock_30m_kdata import __all__ as _stock_30m_kdata_all
__all__ += _stock_30m_kdata_all


# the __all__ is generated
__all__ = []

# __init__.py structure:
# common code of the package
# export interface in __all__ which contains __all__ of its sub modules

# import all from submodule eastmoney_block_meta_recorder
from .eastmoney_block_meta_recorder import *
from .eastmoney_block_meta_recorder import __all__ as _eastmoney_block_meta_recorder_all
__all__ += _eastmoney_block_meta_recorder_all

# import all from submodule eastmoney_stock_meta_recorder
from .eastmoney_stock_meta_recorder import *
from .eastmoney_stock_meta_recorder import __all__ as _eastmoney_stock_meta_recorder_all
__all__ += _eastmoney_stock_meta_recorder_all


# -*- coding: utf-8 -*-
from zvtm.api.utils import to_report_period_type
from zvtm.domain import FinanceFactor
from zvtm.recorders.eastmoney.finance.base_china_stock_finance_recorder import BaseChinaStockFinanceRecorder
from zvtm.utils.time_utils import to_pd_timestamp
from zvtm.utils.utils import add_func_to_value, to_float

finance_factor_map = {
    # 基本每股收益(元)
    "basic_eps": "Epsjb",
    # 扣非每股收益(元)
    "deducted_eps": "Epskcjb",
    # 稀释每股收益(元)
    "diluted_eps": "Epsxs",
    # 每股净资产(元)
    "bps": "Bps",
    # 每股资本公积(元)
    "capital_reserve_ps": "Mgzbgj",
    # 每股未分配利润(元)
    "undistributed_profit_ps": "Mgwfplr",
    # 每股经营现金流(元)
    "op_cash_flow_ps": "Mgjyxjje",
    # 成长能力指标
    #
    # 营业总收入(元)
    "total_op_income": "Totalincome",
    # 毛利润(元)
    "gross_profit": "Grossprofit",
    # 归属净利润(元)
    "net_profit": "Parentnetprofit",
    # 扣非净利润(元)
    "deducted_net_profit": "Bucklenetprofit",
    # 营业总收入同比增长
    "op_income_growth_yoy": "Totalincomeyoy",
    # 归属净利润同比增长
    "net_profit_growth_yoy ": "Parentnetprofityoy",
    # 扣非净利润同比增长
    "deducted_net_profit_growth_yoy": "Bucklenetprofityoy",
    # 营业总收入滚动环比增长
    "op_income_growth_qoq": "Totalincomerelativeratio",
    # 归属净利润滚动环比增长
    "net_profit_growth_qoq": "Parentnetprofitrelativeratio",
    # 扣非净利润滚动环比增长
    "deducted_net_profit_growth_qoq": "Bucklenetprofitrelativeratio",
    # 盈利能力指标
    #
    # 净资产收益率(加权)
    "roe": "Roejq",
    # 净资产收益率(扣非/加权)
    "deducted_roe": "Roekcjq",
    # 总资产收益率(加权)
    "rota": "Allcapitalearningsrate",
    # 毛利率
    "gross_profit_margin": "Grossmargin",
    # 净利率
    "net_margin": "Netinterest",
    # 收益质量指标
    #
    # 预收账款/营业收入
    "advance_receipts_per_op_income": "Accountsrate",
    # 销售净现金流/营业收入
    "sales_net_cash_flow_per_op_income": "Salesrate",
    # 经营净现金流/营业收入
    "op_net_cash_flow_per_op_income": "Operatingrate",
    # 实际税率
    "actual_tax_rate": "Taxrate",
    # 财务风险指标
    #
    # 流动比率
    "current_ratio": "Liquidityratio",
    # 速动比率
    "quick_ratio": "Quickratio",
    # 现金流量比率
    "cash_flow_ratio": "Cashflowratio",
    # 资产负债率
    "debt_asset_ratio": "Assetliabilityratio",
    # 权益乘数
    "em": "Equitymultiplier",
    # 产权比率
    "equity_ratio": "Equityratio",
    # 营运能力指标(一般企业)
    #
    # 总资产周转天数(天)
    "total_assets_turnover_days": "Totalassetsdays",
    # 存货周转天数(天)
    "inventory_turnover_days": "Inventorydays",
    # 应收账款周转天数(天)
    "receivables_turnover_days": "Accountsreceivabledays",
    # 总资产周转率(次)
    "total_assets_turnover": "Totalassetrate",
    # 存货周转率(次)
    "inventory_turnover": "Inventoryrate",
    # 应收账款周转率(次)
    "receivables_turnover": "Accountsreceiveablerate",

    # 专项指标(银行)
    #
    # 存款总额
    "fi_total_deposit": "Totaldeposit",
    # 贷款总额
    "fi_total_loan": "Totalloan",
    # 存贷款比例
    "fi_loan_deposit_ratio": "Depositloanratio",
    # 资本充足率
    "fi_capital_adequacy_ratio": "Capitaladequacyratio",
    # 核心资本充足率
    "fi_core_capital_adequacy_ratio": "Corecapitaladequacyratio",
    # 不良贷款率
    "fi_npl_ratio": "Nplratio",
    # 不良贷款拨备覆盖率
    "fi_npl_provision_coverage": "Nplprovisioncoverage",
    # 资本净额
    "fi_net_capital": "Netcapital_b",
    # 专项指标(保险)
    #
    # 总投资收益率
    "insurance_roi": "Tror",
    # 净投资收益率
    "insurance_net_investment_yield": "Nror",
    # 已赚保费
    "insurance_earned_premium": "Eapre",
    # 赔付支出
    "insurance_payout": "Comexpend",
    # 退保率
    "insurance_surrender_rate": "Surrate",
    # 偿付能力充足率
    "insurance_solvency_adequacy_ratio": "Solvenra",
    # 专项指标(券商)
    #
    # 净资本
    "broker_net_capital": "Netcapital",
    # 净资产
    "broker_net_assets": "Netassets",
    # 净资本/净资产
    "broker_net_capital_assets_ratio": "Captialrate",
    # 自营固定收益类证券规模/净资本
    "broker_self_operated_fixed_income_securities_net_capital_ratio": "Incomesizerate",
}

add_func_to_value(finance_factor_map, to_float)
finance_factor_map["report_period"] = ("ReportDate", to_report_period_type)
finance_factor_map["report_date"] = ("ReportDate", to_pd_timestamp)


class ChinaStockFinanceFactorRecorder(BaseChinaStockFinanceRecorder):
    url = 'https://emh5.eastmoney.com/api/CaiWuFenXi/GetZhuYaoZhiBiaoList'
    finance_report_type = 'ZhuYaoZhiBiaoList'

    data_schema = FinanceFactor
    data_type = 1

    def get_data_map(self):
        return finance_factor_map


if __name__ == '__main__':
    # init_log('finance_factor.log')
    recorder = ChinaStockFinanceFactorRecorder(codes=['000001'])
    recorder.run()
# the __all__ is generated
__all__ = ['ChinaStockFinanceFactorRecorder']


# -*- coding: utf-8 -*-
from zvtm.api.utils import to_report_period_type
from zvtm.domain import BalanceSheet
from zvtm.recorders.eastmoney.finance.base_china_stock_finance_recorder import BaseChinaStockFinanceRecorder
from zvtm.utils.time_utils import to_pd_timestamp
from zvtm.utils.utils import add_func_to_value, first_item_to_float

balance_sheet_map = {
    # 流动资产
    #
    # 货币资金
    "cash_and_cash_equivalents": "Monetaryfund",
    # 应收票据
    "note_receivable": "Billrec",
    # 应收账款
    "accounts_receivable": "Accountrec",
    # 预付款项
    "advances_to_suppliers": "Advancepay",
    # 其他应收款
    "other_receivables": "Otherrec",
    # 存货
    "inventories": "Inventory",
    # 一年内到期的非流动资产
    "current_portion_of_non_current_assets": "Nonlassetoneyear",
    # 其他流动资产
    "other_current_assets": "Otherlasset",
    # 流动资产合计
    "total_current_assets": "Sumlasset",
    # 非流动资产
    #
    # 可供出售金融资产
    "fi_assets_saleable": "Saleablefasset",
    # 长期应收款
    "long_term_receivables": "Ltrec",
    # 长期股权投资
    "long_term_equity_investment": "Ltequityinv",
    # 投资性房地产
    "real_estate_investment": "Estateinvest",
    # 固定资产
    "fixed_assets": "Fixedasset",
    # 在建工程
    "construction_in_process": "Constructionprogress",
    # 无形资产
    "intangible_assets": "Intangibleasset",
    # 商誉
    "goodwill": "Goodwill",
    # 长期待摊费用
    "long_term_prepaid_expenses": "Ltdeferasset",
    # 递延所得税资产
    "deferred_tax_assets": "Deferincometaxasset",
    # 其他非流动资产
    "other_non_current_assets": "Othernonlasset",
    # 非流动资产合计
    "total_non_current_assets": "Sumnonlasset",
    # 资产总计
    "total_assets": "Sumasset",
    # 流动负债
    #
    # 短期借款
    "short_term_borrowing": "Stborrow",
    # 吸收存款及同业存放
    "accept_money_deposits": "Deposit",
    # 应付账款
    "accounts_payable": "Accountpay",
    # 预收款项
    "advances_from_customers": "Advancereceive",
    # 应付职工薪酬
    "employee_benefits_payable": "Salarypay",
    # 应交税费
    "taxes_payable": "Taxpay",
    # 应付利息
    "interest_payable": "Interestpay",
    # 其他应付款
    "other_payable": "Otherpay",
    # 一年内到期的非流动负债
    "current_portion_of_non_current_liabilities": "Nonlliaboneyear",
    # 其他流动负债
    "other_current_liabilities": "Otherlliab",
    # 流动负债合计
    "total_current_liabilities": "Sumlliab",
    # 非流动负债
    #
    # 长期借款
    "long_term_borrowing": "Ltborrow",
    # 长期应付款
    "long_term_payable": "Ltaccountpay",
    # 递延收益
    "deferred_revenue": "Deferincome",
    # 递延所得税负债
    "deferred_tax_liabilities": "Deferincometaxliab",
    # 其他非流动负债
    "other_non_current_liabilities": "Othernonlliab",
    # 非流动负债合计
    "total_non_current_liabilities": "Sumnonlliab",
    # 负债合计
    "total_liabilities": "Sumliab",
    # 所有者权益(或股东权益)
    #
    # 实收资本（或股本）
    "capital": "Sharecapital",
    # 资本公积
    "capital_reserve": "Capitalreserve",
    # 专项储备
    "special_reserve": "Specialreserve",
    # 盈余公积
    "surplus_reserve": "Surplusreserve",
    # 未分配利润
    "undistributed_profits": "Retainedearning",
    # 归属于母公司股东权益合计
    "equity": "Sumparentequity",
    # 少数股东权益
    "equity_as_minority_interest": "Minorityequity",
    # 股东权益合计
    "total_equity": "Sumshequity",
    # 负债和股东权益合计
    "total_liabilities_and_equity": "Sumliabshequity",

    # 银行相关
    # 资产
    # 现金及存放中央银行款项
    "fi_cash_and_deposit_in_central_bank": "Cashanddepositcbank",
    # 存放同业款项
    "fi_deposit_in_other_fi": "Depositinfi",
    # 贵金属
    "fi_expensive_metals": "Preciousmetal",
    # 拆出资金
    "fi_lending_to_other_fi": "Lendfund",
    # 以公允价值计量且其变动计入当期损益的金融资产
    "fi_financial_assets_effect_current_income": "Fvaluefasset",
    # 衍生金融资产
    "fi_financial_derivative_asset": "Derivefasset",
    # 买入返售金融资产
    "fi_buying_sell_back_fi__asset": "Buysellbackfasset",
    # 应收账款
    #
    # 应收利息
    "fi_interest_receivable": "Interestrec",
    # 发放贷款及垫款
    "fi_disbursing_loans_and_advances": "Loanadvances",
    # 可供出售金融资产
    #
    # 持有至到期投资
    "fi_held_to_maturity_investment": "Heldmaturityinv",
    # 应收款项类投资
    "fi_account_receivable_investment": "Investrec",
    # 投资性房地产
    #
    # 固定资产
    #
    # 无形资产
    #
    # 商誉
    #
    # 递延所得税资产
    #
    # 其他资产
    "fi_other_asset": "Otherasset",
    # 资产总计
    #
    # 负债
    #
    # 向中央银行借款
    "fi_borrowings_from_central_bank": "Borrowfromcbank",
    # 同业和其他金融机构存放款项
    "fi_deposit_from_other_fi": "Fideposit",
    # 拆入资金
    "fi_borrowings_from_fi": "Borrowfund",
    # 以公允价值计量且其变动计入当期损益的金融负债
    "fi_financial_liability_effect_current_income": "Fvaluefliab",
    # 衍生金融负债
    "fi_financial_derivative_liability": "Derivefliab",
    # 卖出回购金融资产款
    "fi_sell_buy_back_fi_asset": "Sellbuybackfasset",
    # 吸收存款
    "fi_savings_absorption": "Acceptdeposit",
    # 存款证及应付票据
    "fi_notes_payable": "Cdandbillrec",
    # 应付职工薪酬
    #
    # 应交税费
    #
    # 应付利息
    #
    # 预计负债
    "fi_estimated_liabilities": "Anticipateliab",
    # 应付债券
    "fi_bond_payable": "Bondpay",
    # 其他负债
    "fi_other_liability": "Otherliab",
    # 负债合计
    #
    # 所有者权益(或股东权益)
    # 股本
    "fi_capital": "Shequity",
    # 其他权益工具
    "fi_other_equity_instruments": "Otherequity",
    # 其中:优先股
    "fi_preferred_stock": "Preferredstock",
    # 资本公积
    #
    # 盈余公积
    #
    # 一般风险准备
    "fi_generic_risk_reserve": "Generalriskprepare",
    # 未分配利润
    #
    # 归属于母公司股东权益合计
    #
    # 股东权益合计
    #
    # 负债及股东权益总计

    # 券商相关
    # 资产
    #
    # 货币资金
    #
    # 其中: 客户资金存款
    "fi_client_fund": "Clientfund",
    # 结算备付金
    "fi_deposit_reservation_for_balance": "Settlementprovision",
    # 其中: 客户备付金
    "fi_client_deposit_reservation_for_balance": "Clientprovision",
    # 融出资金
    "fi_margin_out_fund": "Marginoutfund",
    # 以公允价值计量且其变动计入当期损益的金融资产
    #
    # 衍生金融资产
    #
    # 买入返售金融资产
    #
    # 应收利息
    #
    # 应收款项
    "fi_receivables": "Receivables",
    # 存出保证金
    "fi_deposit_for_recognizance": "Gdepositpay",
    # 可供出售金融资产
    #
    # 持有至到期投资
    #
    # 长期股权投资
    #
    # 固定资产
    #
    # 在建工程
    #
    # 无形资产
    #
    # 商誉
    #
    # 递延所得税资产
    #
    # 其他资产
    #
    # 资产总计
    #
    # 负债
    #
    # 短期借款
    #
    # 拆入资金
    #
    # 以公允价值计量且其变动计入当期损益的金融负债
    #
    # 衍生金融负债
    #
    # 卖出回购金融资产款
    #
    # 代理买卖证券款
    "fi_receiving_as_agent": "Agenttradesecurity",
    # 应付账款
    #
    # 应付职工薪酬
    #
    # 应交税费
    #
    # 应付利息
    #
    # 应付短期融资款
    "fi_short_financing_payable": "Shortfinancing",
    # 预计负债
    #
    # 应付债券
    #
    # 递延所得税负债
    #
    # 其他负债
    #
    # 负债合计
    #
    # 所有者权益(或股东权益)
    #
    # 股本
    #
    # 资本公积
    #
    # 其他权益工具
    #
    # 盈余公积
    #
    # 一般风险准备
    #
    # 交易风险准备
    "fi_trade_risk_reserve": "Traderiskprepare",
    # 未分配利润
    #
    # 归属于母公司股东权益合计
    #
    # 少数股东权益
    #
    # 股东权益合计
    #
    # 负债和股东权益总计

    # 保险相关
    # 应收保费
    "fi_premiums_receivable": "Premiumrec",
    "fi_reinsurance_premium_receivable": "Rirec",
    # 应收分保合同准备金
    "fi_reinsurance_contract_reserve": "Ricontactreserverec",
    # 保户质押贷款
    "fi_policy_pledge_loans": "Insuredpledgeloan",
    # 定期存款
    "fi_time_deposit": "Tdeposit",
    # 可供出售金融资产
    #
    # 持有至到期投资
    #
    # 应收款项类投资
    #
    # 应收账款
    #
    # 长期股权投资
    #
    # 存出资本保证金
    "fi_deposit_for_capital_recognizance": "Capitalgdepositpay",
    # 投资性房地产
    #
    # 固定资产
    #
    # 无形资产
    #
    # 商誉
    #
    # 递延所得税资产
    #
    # 其他资产
    #
    # 独立账户资产
    "fi_capital_in_independent_accounts": "Independentasset",
    # 资产总计
    #
    # 负债
    #
    # 短期借款
    #
    # 同业及其他金融机构存放款项
    #
    # 拆入资金
    #
    # 以公允价值计量且其变动计入当期损益的金融负债
    #
    # 衍生金融负债
    #
    # 卖出回购金融资产款
    #
    # 吸收存款
    #
    # 代理买卖证券款
    #
    # 应付账款
    #
    # 预收账款
    "fi_advance_from_customers": "Advancerec",
    # 预收保费
    "fi_advance_premium": "Premiumadvance",
    # 应付手续费及佣金
    "fi_fees_and_commissions_payable": "Commpay",
    # 应付分保账款
    "fi_dividend_payable_for_reinsurance": "Ripay",
    # 应付职工薪酬
    #
    # 应交税费
    #
    # 应付利息
    #
    # 预计负债
    #
    # 应付赔付款
    "fi_claims_payable": "Claimpay",
    # 应付保单红利
    "fi_policy_holder_dividend_payable": "Policydivipay",
    # 保户储金及投资款
    "fi_policy_holder_deposits_and_investment_funds": "Insureddepositinv",
    # 保险合同准备金
    "fi_contract_reserve": "Contactreserve",
    # 长期借款
    #
    # 应付债券
    #
    # 递延所得税负债
    #
    # 其他负债
    #
    # 独立账户负债
    "fi_independent_liability": "Independentliab",
    # 负债合计
    #
    # 所有者权益(或股东权益)
    #
    # 股本
    #
    # 资本公积
    #
    # 盈余公积
    #
    # 一般风险准备
    #
    # 未分配利润
    #
    # 归属于母公司股东权益总计
    #
    # 少数股东权益
    #
    # 股东权益合计
    #
    # 负债和股东权益总计

}

add_func_to_value(balance_sheet_map, first_item_to_float)
balance_sheet_map["report_period"] = ("ReportDate", to_report_period_type)
balance_sheet_map["report_date"] = ("ReportDate", to_pd_timestamp)


class ChinaStockBalanceSheetRecorder(BaseChinaStockFinanceRecorder):
    data_schema = BalanceSheet

    url = 'https://emh5.eastmoney.com/api/CaiWuFenXi/GetZiChanFuZhaiBiaoList'
    finance_report_type = 'ZiChanFuZhaiBiaoList'
    data_type = 3

    def get_data_map(self):
        return balance_sheet_map


if __name__ == '__main__':
    # init_log('blance_sheet.log')
    recorder = ChinaStockBalanceSheetRecorder(codes=['002572'])
    recorder.run()
# the __all__ is generated
__all__ = ['ChinaStockBalanceSheetRecorder']


# -*- coding: utf-8 -*-

import pandas as pd

from zvtm.contract.factor import Scorer, Transformer
from zvtm.utils.pd_utils import normal_index_df


def ma(s: pd.Series, window: int = 5):
    """

    :param s:
    :param window:
    :return:
    """
    return s.rolling(window=window, min_periods=window).mean()


def ema(s, window=12):
    return s.ewm(span=window, adjust=False, min_periods=window).mean()


def live_or_dead(x):
    if x:
        return 1
    else:
        return -1


def macd(s, slow=26, fast=12, n=9, return_type='df', normal=False, count_live_dead=False):
    # 短期均线
    ema_fast = ema(s, window=fast)
    # 长期均线
    ema_slow = ema(s, window=slow)

    # 短期均线 - 长期均线 = 趋势的力度
    diff = ema_fast - ema_slow
    # 力度均线
    dea = diff.ewm(span=n, adjust=False).mean()

    # 力度 的变化
    m = (diff - dea) * 2

    # normal it
    if normal:
        diff = diff / s
        dea = dea / s
        m = m / s

    if count_live_dead:
        live = (diff > dea).apply(lambda x: live_or_dead(x))
        bull = (diff > 0) & (dea > 0)
        live_count = live * (live.groupby((live != live.shift()).cumsum()).cumcount() + 1)

    if return_type == 'se':
        if count_live_dead:
            return diff, dea, m, live, bull, live_count
        return diff, dea, m
    else:
        if count_live_dead:
            return pd.DataFrame(
                {'diff': diff, 'dea': dea, 'macd': m, 'live': live, 'bull': bull, 'live_count': live_count})
        return pd.DataFrame({'diff': diff, 'dea': dea, 'macd': m})


def point_in_range(point: float, range: tuple):
    """

    :param point: one point
    :param range: (start,end)
    :return:
    """
    return range[0] <= point <= range[1]


def intersect_ranges(range_list):
    if len(range_list) == 1:
        return range_list[0]

    result = intersect(range_list[0], range_list[1])
    for range_i in range_list[2:]:
        result = intersect(result, range_i)
    return result


def combine(range_a, range_b):
    if intersect(range_a, range_b):
        return min(range_a[0], range_b[0]), max(range_a[1], range_b[1])
    return None


def intersect(range_a, range_b):
    """
    range_a and range_b with format (start,end) in y axis

    :param range_a:
    :param range_b:
    :return:
    """
    if not range_a or not range_b:
        return None
    # 包含
    if point_in_range(range_a[0], range_b) and point_in_range(range_a[1], range_b):
        return range_a
    if point_in_range(range_b[0], range_a) and point_in_range(range_b[1], range_a):
        return range_b

    if point_in_range(range_a[0], range_b):
        return range_a[0], range_b[1]

    if point_in_range(range_b[0], range_a):
        return range_b[0], range_a[1]
    return None


class RankScorer(Scorer):
    def __init__(self, ascending=True) -> None:
        self.ascending = ascending

    def score(self, input_df) -> pd.DataFrame:
        result_df = input_df.groupby(level=1).rank(ascending=self.ascending, pct=True)
        return result_df


class MaTransformer(Transformer):
    def __init__(self, windows=[5, 10], cal_change_pct=False) -> None:
        super().__init__()
        self.windows = windows
        self.cal_change_pct = cal_change_pct

    def transform_one(self, entity_id, df: pd.DataFrame) -> pd.DataFrame:
        if self.cal_change_pct:
            df['change_pct'] = df['close'].pct_change()

        for window in self.windows:
            col = 'ma{}'.format(window)
            self.indicators.append(col)

            df[col] = df['close'].rolling(window=window, min_periods=window).mean()

        return df


class IntersectTransformer(Transformer):
    def __init__(self, kdata_overlap=0) -> None:
        super().__init__()
        self.kdata_overlap = kdata_overlap

    def transform(self, input_df) -> pd.DataFrame:
        if self.kdata_overlap > 0:
            # 没有重叠，区间就是(0,0)
            input_df['overlap'] = [(0, 0)] * len(input_df.index)

            def cal_overlap(s):
                high = input_df.loc[s.index, 'high']
                low = input_df.loc[s.index, 'low']
                intersection = intersect_ranges(list(zip(low.to_list(), high.to_list())))
                if intersection:
                    # 设置column overlap为intersection,即重叠区间
                    input_df.at[s.index[-1], 'overlap'] = intersection
                return 0

            input_df[['high', 'low']].groupby(level=0).rolling(window=self.kdata_overlap,
                                                               min_periods=self.kdata_overlap).apply(
                cal_overlap, raw=False)

        return input_df


class MaAndVolumeTransformer(Transformer):
    def __init__(self, windows=[5, 10], vol_windows=[30], kdata_overlap=0) -> None:
        super().__init__()
        self.windows = windows
        self.vol_windows = vol_windows
        self.kdata_overlap = kdata_overlap

    def transform(self, input_df) -> pd.DataFrame:
        for window in self.windows:
            col = 'ma{}'.format(window)
            self.indicators.append(col)

            ma_df = input_df['close'].groupby(level=0).rolling(window=window, min_periods=window).mean()
            ma_df = ma_df.reset_index(level=0, drop=True)
            input_df[col] = ma_df

        for vol_window in self.vol_windows:
            col = 'vol_ma{}'.format(vol_window)
            self.indicators.append(col)

            vol_ma_df = input_df['volume'].groupby(level=0).rolling(window=vol_window, min_periods=vol_window).mean()
            vol_ma_df = vol_ma_df.reset_index(level=0, drop=True)
            input_df[col] = vol_ma_df

        if self.kdata_overlap > 0:
            input_df['overlap'] = [(0, 0)] * len(input_df.index)

            def cal_overlap(s):
                high = input_df.loc[s.index, 'high']
                low = input_df.loc[s.index, 'low']
                intersection = intersect_ranges(list(zip(low.to_list(), high.to_list())))
                if intersection:
                    input_df.at[s.index[-1], 'overlap'] = intersection
                return 0

            input_df[['high', 'low']].groupby(level=0).rolling(window=self.kdata_overlap,
                                                               min_periods=self.kdata_overlap).apply(
                cal_overlap, raw=False)

        return input_df


class MacdTransformer(Transformer):
    def __init__(self, slow=26, fast=12, n=9, normal=False, count_live_dead=False) -> None:
        super().__init__()
        self.slow = slow
        self.fast = fast
        self.n = n
        self.normal = normal
        self.count_live_dead = count_live_dead

        self.indicators.append('diff')
        self.indicators.append('dea')
        self.indicators.append('macd')

    def transform(self, input_df) -> pd.DataFrame:
        macd_df = input_df.groupby(level=0)['close'].apply(
            lambda x: macd(x, slow=self.slow, fast=self.fast, n=self.n, return_type='df', normal=self.normal,
                           count_live_dead=self.count_live_dead))
        input_df = pd.concat([input_df, macd_df], axis=1, sort=False)
        return input_df

    def transform_one(self, entity_id, df: pd.DataFrame) -> pd.DataFrame:
        print(f'transform_one {entity_id} {df}')
        return macd(df['close'], slow=self.slow, fast=self.fast, n=self.n, return_type='df', normal=self.normal,
                    count_live_dead=self.count_live_dead)


class QuantileScorer(Scorer):
    def __init__(self, score_levels=[0, 0.1, 0.3, 0.5, 0.7, 0.9, 1.0]) -> None:
        self.score_levels = score_levels

    def score(self, input_df):
        self.score_levels.sort(reverse=True)

        quantile_df = input_df.groupby(level=1).quantile(self.score_levels)
        quantile_df.index.names = [self.time_field, 'score']

        self.logger.info('factor:{},quantile:\n{}'.format(self.factor_name, quantile_df))

        result_df = input_df.copy()
        result_df.reset_index(inplace=True, level='entity_id')
        result_df['quantile'] = None
        for timestamp in quantile_df.index.levels[0]:
            length = len(result_df.loc[result_df.index == timestamp, 'quantile'])
            result_df.loc[result_df.index == timestamp, 'quantile'] = [quantile_df.loc[
                                                                           timestamp].to_dict()] * length

        self.logger.info('factor:{},df with quantile:\n{}'.format(self.factor_name, result_df))

        # result_df = result_df.set_index(['entity_id'], append=True)
        # result_df = result_df.sort_index(level=[0, 1])
        #
        # self.logger.info(result_df)
        #
        def calculate_score(df, factor_name, quantile):
            original_value = df[factor_name]
            score_map = quantile.get(factor_name)
            min_score = self.score_levels[-1]

            if original_value < score_map.get(min_score):
                return 0

            for score in self.score_levels[:-1]:
                if original_value >= score_map.get(score):
                    return score

        for factor in input_df.columns.to_list():
            result_df[factor] = result_df.apply(lambda x: calculate_score(x, factor, x['quantile']),
                                                axis=1)

        result_df = result_df.reset_index()
        result_df = normal_index_df(result_df)
        result_df = result_df.loc[:, self.factors]

        result_df = result_df.loc[~result_df.index.duplicated(keep='first')]

        self.logger.info('factor:{},df:\n{}'.format(self.factor_name, result_df))

        return result_df


# the __all__ is generated
__all__ = ['ma', 'ema', 'live_or_dead', 'macd', 'point_in_range', 'intersect_ranges', 'intersect', 'RankScorer',
           'MaTransformer', 'IntersectTransformer', 'MaAndVolumeTransformer', 'MacdTransformer', 'QuantileScorer']



# -*- coding: utf-8 -*-
import logging
from typing import List

import sqlalchemy
from sqlalchemy.ext.declarative import DeclarativeMeta

from zvtm.contract import TradableEntity, zvt_context, Mixin
from zvtm.contract.api import get_db_engine, get_db_session_factory
from zvtm.utils.utils import add_to_map_list

logger = logging.getLogger(__name__)


def register_entity(entity_type: str = None):
    """
    function for register entity type

    :param entity_type:
    :type entity_type:
    :return:
    :rtype:
    """

    def register(cls):
        # register the entity
        if issubclass(cls, TradableEntity):
            entity_type_ = entity_type
            if not entity_type:
                entity_type_ = cls.__name__.lower()

            if entity_type_ not in zvt_context.tradable_entity_types:
                zvt_context.tradable_entity_types.append(entity_type_)
                zvt_context.entity_schemas.append(cls)
            zvt_context.tradable_schema_map[entity_type_] = cls
        return cls

    return register


def register_schema(providers: List[str],
                    db_name: str,
                    schema_base: DeclarativeMeta,
                    entity_type: str = None):
    """
    function for register schema,please declare them before register

    :param providers: the supported providers for the schema
    :type providers:
    :param db_name: database name for the schema
    :type db_name:
    :param schema_base:
    :type schema_base:
    :param entity_type: the schema related entity_type
    :type entity_type:
    :return:
    :rtype:
    """
    schemas = []
    for item in schema_base.registry.mappers:
        cls = item.class_
        if type(cls) == DeclarativeMeta:
            # register provider to the schema
            for provider in providers:
                if issubclass(cls, Mixin):
                    cls.register_provider(provider)

            if zvt_context.dbname_map_schemas.get(db_name):
                schemas = zvt_context.dbname_map_schemas[db_name]
            zvt_context.schemas.append(cls)
            if entity_type:
                add_to_map_list(the_map=zvt_context.entity_map_schemas, key=entity_type, value=cls)
            schemas.append(cls)

    zvt_context.dbname_map_schemas[db_name] = schemas

    for provider in providers:
        if provider not in zvt_context.providers:
            zvt_context.providers.append(provider)

        if not zvt_context.provider_map_dbnames.get(provider):
            zvt_context.provider_map_dbnames[provider] = []
        zvt_context.provider_map_dbnames[provider].append(db_name)
        zvt_context.dbname_map_base[db_name] = schema_base

        # create the db & table
        engine = get_db_engine(provider, db_name=db_name)
        schema_base.metadata.create_all(engine)

        session_fac = get_db_session_factory(provider, db_name=db_name)
        session_fac.configure(bind=engine)

    for provider in providers:
        engine = get_db_engine(provider, db_name=db_name)

        # create index for 'timestamp','entity_id','code','report_period','updated_timestamp
        for table_name, table in iter(schema_base.metadata.tables.items()):
            index_list = []
            with engine.connect() as con:
                #rs = con.execute("PRAGMA INDEX_LIST('{}')".format(table_name))
                rs = con.execute("show index from {}".format(table_name))
                for row in rs:
                    index_list.append(row[2])

            logger.debug('engine:{},table:{},index:{}'.format(engine, table_name, index_list))

            for col in ['timestamp', 'entity_id', 'code', 'report_period', 'created_timestamp', 'updated_timestamp']:
                if col in table.c:
                    column = eval('table.c.{}'.format(col))
                    index_name = '{}_{}_index'.format(table_name, col)
                    if index_name not in index_list:
                        index = sqlalchemy.schema.Index(index_name, column)
                        index.create(engine)
            for cols in [('timestamp', 'entity_id'), ('timestamp', 'code')]:
                if (cols[0] in table.c) and (col[1] in table.c):
                    column0 = eval('table.c.{}'.format(col[0]))
                    column1 = eval('table.c.{}'.format(col[1]))
                    index_name = '{}_{}_{}_index'.format(table_name, col[0], col[1])
                    if index_name not in index_list:
                        index = sqlalchemy.schema.Index(index_name, column0,
                                                        column1)
                        index.create(engine)
# the __all__ is generated
__all__ = ['register_entity', 'register_schema']


# -*- coding: utf-8 -*-

import io

import pandas as pd
import requests

from zvtm.api.utils import china_stock_code_to_id
from zvtm.contract.api import df_to_db
from zvtm.contract.recorder import Recorder, TimeSeriesDataRecorder
from zvtm.domain import Index, IndexCategory, IndexStock
from zvtm.recorders.consts import DEFAULT_HEADER
from zvtm.utils.time_utils import to_pd_timestamp


# 上证指数 中证指数
class ExchangeCSIndexMetaRecorder(Recorder):
    original_page_url = 'http://www.csindex.com.cn/zh-CN/indices/index?class_2=2&class_17=17&class_10=10&class_7=7&is_custom_0=1'

    data_schema = Index
    provider = 'exchange'

    url = 'http://www.csindex.com.cn/zh-CN/indices/index?page=1&page_size=1000&by=asc&order=%E5%8F%91%E5%B8%83%E6%97%B6%E9%97%B4&data_type=json' \
          '&{}&class_7=7&class_10=10&{}&is_custom_0=1'

    # 中证指数 抓取 风格指数 行业指数 规模指数
    csi_category_map_url = {
        IndexCategory.style: url.format('class_1=1', 'class_19=19'),
        IndexCategory.industry: url.format('class_1=1', 'class_18=18'),
        IndexCategory.scope: url.format('class_1=1', 'class_17=17'),
    }

    # 上证指数 只取规模指数
    sh_category_map_url = {
        IndexCategory.scope: url.format('class_2=2', 'class_17=17')
    }

    def run(self):
        self.record_index('sh')
        self.record_index('csi')

    def get_resp_data(self, resp: requests.Response):
        resp.raise_for_status()
        return resp.json()['list']

    def record_index(self, index_type):
        if index_type == 'csi':
            category_map_url = self.csi_category_map_url
        elif index_type == 'sh':
            category_map_url = self.sh_category_map_url
        else:
            self.logger.warning(f'not support index type: {index_type}')
            assert False

        requests_session = requests.Session()

        for category, url in category_map_url.items():
            resp = requests_session.get(url, headers=DEFAULT_HEADER)

            results = self.get_resp_data(resp)
            the_list = []

            self.logger.info(f'category: {category} ')
            self.logger.info(f'results: {results} ')
            for i, result in enumerate(results):
                self.logger.info(f'to {i}/{len(results)}')
                code = result['index_code']
                name = result['indx_sname']
                entity_id = f'index_sh_{code}'
                index_item = {
                    'id': entity_id,
                    'entity_id': entity_id,
                    'timestamp': to_pd_timestamp(result['base_date']),
                    'entity_type': 'index',
                    'exchange': 'sh',
                    'code': code,
                    'name': name,
                    'category': category.value,
                    'list_date': to_pd_timestamp(result['online_date']),
                    'base_point': result['base_point'],
                    'publisher': 'csindex'
                }
                self.logger.info(index_item)
                the_list.append(index_item)
            if the_list:
                df = pd.DataFrame.from_records(the_list)
                df_to_db(data_schema=self.data_schema, df=df, provider=self.provider,
                         force_update=True)
            self.logger.info(f"finish record {index_type} index:{category.value}")


class ExchangeCSIndexStockRecorder(TimeSeriesDataRecorder):
    entity_provider = 'exchange'
    entity_schema = Index

    provider = 'exchange'
    data_schema = IndexStock

    original_page_url = 'http://www.csindex.com.cn/zh-CN/downloads/indices'
    url = 'http://www.csindex.com.cn/uploads/file/autofile/cons/{}cons.xls'

    def record(self, entity, start, end, size, timestamps):
        url = self.url.format(entity.code)
        response = requests.get(url, headers=DEFAULT_HEADER)
        response.raise_for_status()

        df = pd.read_excel(io.BytesIO(response.content))

        df = df[['日期Date', '成分券代码Constituent Code', '成分券名称Constituent Name']].rename(
            columns={'日期Date': 'timestamp', '成分券代码Constituent Code': 'stock_code',
                     '成分券名称Constituent Name': 'stock_name'})

        index_id = f'index_sh_{entity.code}'
        df['entity_id'] = index_id
        df['entity_type'] = 'index'
        df['exchange'] = 'sh'
        df['code'] = entity.code
        df['name'] = entity.name
        df['stock_id'] = df['stock_code'].apply(lambda x: china_stock_code_to_id(str(x)))
        # id format: {entity_id}_{timestamp}_{stock_id}
        df['id'] = df[['entity_id', 'timestamp', 'stock_id']].apply(lambda x: '_'.join(x.astype(str)), axis=1)
        df['timestamp'] = pd.to_datetime(df['timestamp'])

        df_to_db(data_schema=self.data_schema, df=df, provider=self.provider, force_update=True)


__all__ = ['ExchangeCSIndexMetaRecorder']

if __name__ == '__main__':
    # ExchangeIndexMetaRecorder().run()
    ExchangeCSIndexStockRecorder(codes=['000001']).run()
# the __all__ is generated
__all__ = ['ExchangeCSIndexMetaRecorder', 'ExchangeCSIndexStockRecorder']


# -*- coding: utf-8 -*-
from sqlalchemy import Column, String, DateTime, Float, Boolean, Integer
from sqlalchemy.orm import declarative_base

from zvtm.contract.register import register_schema
from zvtm.contract.schema import TradableMeetActor

StockActorBase = declarative_base()


class StockTopTenFreeHolder(StockActorBase, TradableMeetActor):
    __tablename__ = 'stock_top_ten_free_holder'

    report_period = Column(String(length=32))
    report_date = Column(DateTime)

    # 持股数
    holding_numbers = Column(Float)
    # 持股比例
    holding_ratio = Column(Float)
    # 持股市值
    holding_values = Column(Float)


class StockTopTenHolder(StockActorBase, TradableMeetActor):
    __tablename__ = 'stock_top_ten_holder'

    report_period = Column(String(length=32))
    report_date = Column(DateTime)

    # 持股数
    holding_numbers = Column(Float)
    # 持股比例
    holding_ratio = Column(Float)
    # 持股市值
    holding_values = Column(Float)


class StockInstitutionalInvestorHolder(StockActorBase, TradableMeetActor):
    __tablename__ = 'stock_institutional_investor_holder'

    report_period = Column(String(length=32))
    report_date = Column(DateTime)

    # 持股数
    holding_numbers = Column(Float)
    # 持股比例
    holding_ratio = Column(Float)
    # 持股市值
    holding_values = Column(Float)


class StockActorSummary(StockActorBase, TradableMeetActor):
    __tablename__ = 'stock_actor_summary'
    # tradable code
    code = Column(String(length=64))
    # tradable name
    name = Column(String(length=128))

    report_period = Column(String(length=32))
    report_date = Column(DateTime)

    # 变动比例
    change_ratio = Column(Float)
    # 是否完成
    is_complete = Column(Boolean)
    # 持股市值
    actor_type = Column(String(length=128))
    actor_count = Column(Integer)

    # 持股数
    holding_numbers = Column(Float)
    # 持股比例
    holding_ratio = Column(Float)
    # 持股市值
    holding_values = Column(Float)


register_schema(providers=['em'], db_name='stock_actor', schema_base=StockActorBase, entity_type='stock')

# the __all__ is generated
__all__ = ['StockTopTenFreeHolder', 'StockTopTenHolder', 'StockInstitutionalInvestorHolder', 'StockActorSummary']